# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""Base format stuff (metaclass, classes, etc.)"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

# TODO: NamedTuple for tree_stack

import hashlib
from copy import deepcopy
from functools import reduce
from glob import glob
from imp import find_module, load_module
from itertools import dropwhile, islice
from logging import getLogger
from os import extsep, fdopen, stat, walk
from os.path import basename, commonprefix, dirname, exists, join, sep, splitext
from sys import modules
from time import time

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

from lxml import etree

try:
    from .defaults import HASHALGO
except ImportError:
    HASHALGO = 'md5'
from .error import ClufterError, ClufterPlainError
from .plugin_registry import MetaPlugin, PluginRegistry
from .protocol import Protocol
from .utils import arg2wrapped, args2sgpl, args2tuple, args2unwrapped, \
                   classproperty, \
                   filterdict_keep, \
                   filterdict_pop, \
                   head_tail, \
                   hybridmethod, \
                   immutable, \
                   iterattrs, \
                   isinstanceupto, \
                   popattr, \
                   tuplist
from .utils_2to3 import MimicMeta, basestring, bytes_enc, iter_items, xrange
from .utils_func import bifilter_unpack, foreach
from .utils_lxml import etree_parser_safe
from .utils_prog import ProtectedDict, getenv_namespaced, namever_partition
from .utils_xml import rng_get_start, rng_pivot

log = getLogger(__name__)

DEFAULT_ROOT_DIR = join(dirname(__file__), 'formats')


class FormatError(ClufterError):
    pass

class FormatPlainError(ClufterPlainError, FormatError):
    pass


class formats(PluginRegistry):
    """Format registry (to be used as a metaclass for formats)"""
    def __init__(cls, name, bases, attrs):
        # could be called multiple times but only once per respective
        # __new__ (in plugin_registry) is required, rest would be waste
        # of resources if not harmful due to non-idempotent modifications
        if cls._probes > 1:
            return

        cls._protocols, cls._validators, cls._protocol_attrs = {}, {}, set()
        cls._context = set(popattr(cls, 'context_specs',
                                   attrs.pop('context_specs', ())))
        cls._update_to = OrderedDict()  # to be filled "retrospectively"
        compat = popattr(cls, 'compat_contingency',
                         attrs.pop('compat_contingency', ()))
        cls._update_from = OrderedDict((c.in_format, c) for c in compat)
        # protocols merge: top-down through inheritance
        # (real base class goes last, rest are "convertible siblings")
        for i, base in enumerate(reversed(bases)):
            if i or base is MetaPlugin:  # not a proper base or CompositeFormat
                continue
            cls._protocols.update(getattr(base, '_protocols', {}))
            cls._validators.update(getattr(base, '_validators', {}))
            cls._context.update(getattr(base, '_context', ()))
            cls._protocol_attrs.update(getattr(base, '_protocol_attrs', ()))
            if cls.name in base._update_from:
                base._update_from[cls] = flt = base._update_from.pop(cls.name)
                cls._update_to[base] = flt
        # updated with locally defined proto's (marked by `producing` wrapper)
        specs = popattr(cls, 'validator_specs',
                        attrs.pop('validator_specs', {}))

        for attr, obj in iterattrs(cls):
            if isinstance(obj, Protocol):
                cls._protocol_attrs.add(attr)
                continue
            try:
                protocol = obj._protocol
                delattr(obj, '_protocol')
                cls._protocols[protocol] = cls, attr
                cls._validators[protocol] = obj._validator, None
                delattr(obj, '_validator')
            except AttributeError:
                if not hasattr(obj, '_protocol'):
                    assert attr not in cls._protocols, 'Unexpected override'

        for protocol in cls._protocols:
            newspec = specs.get(protocol, None)
            validator, spec = cls._validators.get(protocol, (None, ''))
            newspec = newspec if newspec is not None else spec
            if validator:
                cls._validators[protocol] = validator, newspec


class _Format(object):
    """Base for configuration formats

    Base principles:
        - protocols: string label denoting how to int-/externalize
          - union of protocols within inheritance hierarchy and
            locally defined ones (prioritized)
            - to define one, add a method decorated with
              `@Format.producing(<proto>)`
            - be default, all such protocols are suitable for both
              int- and externalization, but you can prevent the latter
              context by raising an exception in the method body
        - create format instance = internalize, call = externalize
        - protocols are property of the class, representation of an instance
        - basic inheritance rules apply to formats as well, notably that
          an instance of expected class' subclass is generally acceptable,
          but there are very specific subtleties that should rather be
          spelled out:
          - for a format hierarchy that explicitly arranges for older-to-newer
            instance upgrades (`cib` family of formats is one such example)
            - inheritance for `fmt` placeholder format should be arranged
              like this:  Format <-- ... <-- fmt <-- fmtN <-- ... <-- fmtM
              (also, care to stick with a linear inheritance between formats)
              (where "<--" denotes inheritance as usual, and fmtM/fmtN
              stands for oldest/newest supported version, respectively, hence
              we can conclude that older format versions derive from newer
              ones, transitively)
            - `fmt` top of the formats hierarchy should be the generalization
              of the format, that is, able to accept arbitrary instance of
              its subclasses-or-self (as long as the format has a way to
              distinguish the format version internally so it won't get lost)
            - for the rest (tail) of the hierarchy, we can easily see that
              instances of newer format versions cannot be used where older
              ones are explicitly required (forward-consciousness is not
              assumed at all), but the opposite direction, that is,
              backward-consciousness, is deliberately possible, implicitly
              through trial-and-error (a.k.a. lazy) arrangement (see below)
            - implicit forward compatibility takes the least-resistance
              approach -- no hassles are needed when an instance of an older
              format version validates per requirements imposed on the newer
              format, otherwise this new format is supposed to specify
              filter(s), via its `compat_contingency` member, sufficient
              to promote the older instance so it eventually validates
              (it's a fatal failure if this cannot be achieved)


    Little bit of explanation:

                           FORMAT INSTANCE
                           (concrete data)
     INSTANTIATION    /-----------------------\       CALL        /--------
     (internalize)    | -protocols = set(...) |   (externalize)   | effect
    ----------------->| -representations =    |------------------>| and/or ...
    (protocol, *args) |   {protocol: (*args)} | (protocol, *args) |  value
           ^          \-----------------------/             ^     \--------
           |                                                |
           |                                                |
           +-- adds (protocol, *args) to representations,   |
               usually does nothing else (lazy approach)    |
                                                            |
             with each protocol there is a representation --+
             of concrete data associated,  either this or
             any other  that  can be promoted  (awakening
             from the previous laziness)  to the desired
             one is required

        Externalization methods are marked with `producing` decorator
        that takes protocol name as a parameter.

    """
    context_specs = 'validator_specs', 'compat_contingency'

    @MimicMeta.passdeco(classproperty)
    def context(self):
        return tuple(self._context)

    @MimicMeta.passdeco(classproperty)
    def native_protocol(self):
        """Native protocol"""
        raise AttributeError

    ###

    @MimicMeta.method
    def _swallow(self, protocol, *args):
        """Called by native constructor/producer to store a format instance"""
        if protocol == 'native':
            protocol = self.native_protocol

        assert protocol in self._protocols, ("unrecognized protocol `{0}'"
                                             .format(protocol))
        assert args != (None, )
        constructing = not self._representations  # called from constructor
        prev = self._representations.setdefault(protocol, args)
        assert prev is args

        if not constructing:
            return  # trust that equivalent of once validated keeps the property
        log.debug("Will try to validate `{0}' instance"
                  .format(self.__class__.name))

        cands = (protocol, self.native_protocol) + tuple(self._validators)[:1]
        validating_protocol = args2unwrapped(*islice(dropwhile(lambda x: x not in
                                                               self._validators,
                                                               cands), 1))
        validator = self.validator(validating_protocol)
        if not validator:
            return  # cannot validate in any way

        obj = self(validating_protocol)
        entries, _ = head_tail(validator(obj))
        if isinstance(entries, basestring):
            log.warning(entries)
        elif entries:
            raise FormatError(self, "Validation: {0}".format(
                ', '.join(':'.join(args2tuple(str(e[0]), str(e[1]), *e[2:]))
                          for e in entries))
            )

    @MimicMeta.method
    def producer(self, protocol):
        protocol_cls, protocol_attr = self._protocols[protocol]
        if protocol_cls is self.__class__:
            return getattr(self, protocol_attr)  # has to be self
        return getattr(super(self.__class__, self), protocol_attr)

    @MimicMeta.method
    def produce(self, protocol, *args, **kwargs):
        """Called by implicit invocation to get data externalized"""
        if protocol == 'native':
            protocol = self.native_protocol

        assert protocol in self._protocols
        ret = self.producer(protocol)(protocol, *args, **kwargs)
        if ret is None:
            raise FormatError(self, "Cannot produce `{0}' format"
                                    " as `{1}'"
                                    .format(self.__class__.name, protocol))
        return ret

    @MimicMeta.method
    def __new__(cls, protocol, *args, **kwargs):
        """Format pre-constructor (possibly format-upgrading to parent class)"""
        good, bad = bifilter_unpack(lambda c, f: not isinstance(c, basestring),
                                    iter_items(cls._update_from))
        if bad:
            log.warning("Unresolved update-from filters for `{0}': {1}".format(
                        cls.name, ','.join(c for (c, _) in bad)))
        retself, worklist = None, list(reversed(good)) + [(cls, None)]
        while worklist:
            tryfmt, tryflt = worklist.pop()
            log.debug("Trying pre-construction of `{0}' format going from"
                      " `{1}'{2}"
                      .format(cls.name, tryfmt.name,
                              tryflt.name.join((" (using `", "' filter)"))
                              if tryflt else ""))
            try:
                # for tryflt case, we cannot use `tryflt.in_format` as it is
                # just a string here
                obj = object.__new__(tryfmt)
                obj._construct(protocol, *args, **kwargs)
                if not tryflt:
                    retself = obj
                else:
                    # note that tryflt is filter class that needs to be
                    # instantiated by being passed dictionary to resolved
                    # its formats at
                    flt = tryflt(formats={
                        tryflt.in_format: tryfmt,
                        cls.name: cls,
                    })
                    retself = flt(obj)
            except FormatError as e:
                retself = e
                continue
            else:
                break
        assert retself
        if isinstance(retself, FormatError):
            raise retself
        return retself

    @MimicMeta.method
    def _construct(self, protocol, *args, **kwargs):
        """Format constructor, i.e., object = concrete internal data"""
        # cannot be doubly-undescored otherwise the "foreign lookup" will miss
        assert not(getattr(self, '_representations', None)), "int. API misuse"
        rs = {}
        self._representations, self._representations_ro = rs, ProtectedDict(rs)
        if not hasattr(self, '_hash'):  # can be defined at the class level
            self._hash = None
        validator_specs = kwargs.pop('validator_specs', {})
        default = validator_specs.setdefault('', None)  # None ~ don't track
        validators = {}
        for p in self._validators:
            spec = validator_specs.get(p, default)
            if spec is None:
                continue
            validators[p] = (self._validators[p][0], spec)
        if validators:  # force per-instance customization
            self._validators = dict(self._validators, **validators)
        self._dict = kwargs
        # supercharge protocol objects with callability to provide
        # a syntactic sugar: self(self.PROTO, ...) -> self.PROTO(...))
        for attr in self._protocol_attrs:
            def get_protocol_proxy(obj):
                class wrapped_call(object):
                    #def __getattribute__(this, name):
                    #    return obj.__getattribute__(name)
                    #def __str__(self):
                    #    return str(obj)
                    def __call__(this, *a, **kw):
                        return self(this, *a, **kw)
                    def __eq__(self, other):
                        return str(obj) == other
                    def __hash__(self):  # for "x in bars"
                        return hash(obj)
                return wrapped_call()
            log.debug("Proxying {0} to add callability".format(attr))
            setattr(self, attr, get_protocol_proxy(getattr(self, attr)))
        self._swallow(protocol, *args)

    @MimicMeta.passdeco(classmethod)
    def common_protocols(cls, foreign):
        """Get common local vs. foreign protocols, native ones prioritized"""
        # XXX composite format
        if not hasattr(foreign, '_protocols'):
            raise FormatError(cls, "cannot check common protocols for `{0}'",
                              repr(foreign))
        return sorted(
            reduce(
                set.intersection,
                map(set, (cls._protocols, foreign._protocols))
            ),
            key=lambda x:
                int(x == cls.native_protocol) * 2
                + int(x == foreign.native_protocol)
        )

    @MimicMeta.classmethod
    def as_instance(cls, *decl_or_instance, **kwargs):
        """Create an instance or verify and return existing one"""
        if decl_or_instance and isinstance(decl_or_instance[0], Format):
            instance = decl_or_instance[0]
            if not isinstanceupto(instance, cls, Format):
                raise FormatError(cls, "input object: format mismatch"
                                  " (expected `{0}' or a superclass, got"
                                  " `{1}')", cls.name, instance.__class__.name)
            if instance.__class__.__bases__[0] is cls:
                return instance

            # convert using (preferably native) protocol
            com = instance.common_protocols(cls)
            if not com:
                raise FormatError(cls, "no common protocol with source format"
                                       " `{0}'", instance.name)
            decl_or_instance = (com[0], instance(com[0]))

        instance = cls(*decl_or_instance, **kwargs)
        return instance

    @MimicMeta.method
    def __call__(self, protocol='native', *args, **kwargs):
        """Way to externalize object's internal data"""
        return self.produce(protocol, *args, **kwargs)

    @MimicMeta.passdeco(classproperty)
    def protocols(self):
        """Set of supported protocols for int-/externalization"""
        return self._protocols.keys()  # installed by meta-level

    @MimicMeta.passdeco(hybridmethod)
    def validator(this, protocol=None, spec=None):
        """Return validating function or None

        This ought to be the authoritative (and only) way to use
        a validator, do not touch _validators and also validator_specs
        is dropped from class attributes for this very reason.
        """
        which = this.native_protocol if protocol is None else protocol
        try:
            validator, sp = this._validators[which]  # installed by meta-level
        except KeyError:
            return None
        spec = spec if spec is not None else sp
        if spec == '':
            return None
        return lambda *args, **kwargs: validator(this, *args,
                                                 **dict(kwargs, spec=spec))

    @MimicMeta.passdeco(property)
    def hash(self):
        if self._hash is None:
            raise NotImplementedError
        return self._hash

    @MimicMeta.passdeco(property)
    def representations(self):
        """Mapping of `protocol: initializing_data`"""
        # XXX should be ProtectedDict
        return self._representations_ro

    @MimicMeta.staticmethod
    def producing(protocol, chained=False, protect=False, validator=None):
        """Decorator for externalizing method understood by the `Format` magic

        As a bonus: caching of representations."""
        def deco_meth(meth):
            def deco_args(self, protocol, *args, **kwargs):
                # XXX enforce nochain for this iterative processing?
                do_protect = protect and not kwargs.pop('protect_safe', False)
                produced = None
                try:
                    # stored -> computed norm.: detuple if len == 1
                    produced = args2unwrapped(*self._representations[protocol])
                except KeyError:
                    produced = None
                    # seemingly absurd inversion of starting with this `prev`:
                    # arranged like this for empty _protocols (hence `get`s)
                    worklist = [(Format, self.__class__)] * 2
                    this_proto = self._protocols[protocol]
                    while worklist:
                        prev_cls, that_cls = worklist.pop()
                        that_proto = that_cls._protocols[protocol]
                        prev_proto = prev_cls._protocols.get(protocol)
                        if that_proto in (this_proto, prev_proto):
                            if worklist or that_proto == prev_proto:
                                if chained:
                                    worklist.extend((that_cls, b) for b in
                                                    that_cls.__bases__ if
                                                    protocol in b._protocols)
                                 # also ensures meth fired atmost once (@level?)
                                continue
                            producer = lambda *args, **kwargs: \
                                           meth(self, *args, **kwargs)
                        else:
                            producer = getattr(super(prev_cls, self),
                                               that_proto[1])
                        produced = producer(protocol, *args, **kwargs)
                        if produced is None:
                            continue
                        if (that_cls is self.__class__
                            and protocol not in self._representations):
                            # computed -> stored normalization
                            self._swallow(protocol, *arg2wrapped(produced))
                        else:
                            do_protect = False
                        break

                if do_protect and not immutable(produced):
                    log.debug("{0}:{1}:Forced deepcopy of `{2}' instance"
                              .format(self.__class__.name, meth.__name__,
                                      type(produced).__name__))
                    produced = deepcopy(produced)
                return produced

            deco_args.__name__, deco_args.__doc__ = meth.__name__, meth.__doc__
            deco_args._protocol = protocol  # mark for later recognition
            if validator:
                deco_args._validator = validator
            return deco_args
        return deco_meth


Format = MimicMeta('Format', formats, _Format)


class Nothing(Format):
    """Empty format, typically an input for "generator" type of filters"""
    native_protocol = VOID = Protocol('void')

    _hash = hash(None)

    @Format.producing(VOID)
    def get_void(self, *iodecl):
        pass  # same as return None


class SimpleFormat(Format):
    """This is what most of the format classes want to subclass"""
    native_protocol = BYTESTRING = Protocol('bytestring')
    FILE = Protocol('file')
    void_file = '/dev/null'  # XXX not multi-platform

    def _construct(self, protocol, *args, **kwargs):
        """Format constructor, i.e., object = concrete uniformat data"""
        assert isinstance(protocol, basestring), \
               "protocol has to be string for `{0}', not `{1}'" \
               .format(self.__class__.name, protocol)
        super(SimpleFormat, self)._construct(protocol, *args, **kwargs)

    @Format.producing(BYTESTRING)
    def get_bytestring(self, *iodecl):
        if self.FILE in self._representations:  # break the possible loop
            infile = self.FILE()
            if hasattr(infile, 'read'):
                # assume fileobj out of our control, do not close
                return bytes_enc(infile.read(), 'utf-8')

            assert isinstance(infile, basestring)
            with open(infile, 'rb') as f:
                return f.read()

    @Format.producing(FILE)
    def get_file(self, *iodecl):
        outfile = iodecl[-1]
        if hasattr(outfile, 'write'):
            # assume fileobj out of our control, do not close
            if hasattr(outfile, 'buffer'):  # PY3: std{in,out,err} text mode
                outfile.buffer.write(self.BYTESTRING())
            else:
                outfile.write(self.BYTESTRING())
            return outfile.name

        assert isinstance(outfile, basestring)
        with open(outfile, 'wb') as f:
            f.write(bytes_enc(self.BYTESTRING(), 'utf-8'))
        return outfile

    @property
    def hash(self):
        """Compute hash trying to uniquely identify the format instance"""
        if self._hash is None:
            # to prevent possible brute-force-based attacks on e.g.,
            # configuration files with sensitive data obfuscated
            # but with hash of the original contained in the output
            # filename as per the common interpolation, we add some
            # salt based either on input's mtime (if it is a file)
            # or current UNIX time (otherwise -- as we cannot anchor
            # it reliably anyway);  this way in such cases, for the
            # unchanged input provided as a file, we always get the
            # identical respective part of the output name, without
            # putting the content of the original at risk (hopefully)
            #
            # manual verification (provided HASHALGO=md5, default):
            # w/o salt:   md5sum $FILE
            # with salt:  { stat --printf "%Y" $FILE; cat $FILE; } | md5sum
            content, salt = '', ''
            hash_algo = getenv_namespaced('HASHALGO', HASHALGO)
            do_salt = getenv_namespaced('NOSALT', '0') in ('0', 'false')
            try:
                hash_algo = getattr(hashlib, hash_algo)
            except AttributeError:
                log.warning("`{0}' hash algorithm unknown".format(hash_algo))
                hash_algo = hashlib.md5
            try:
                h = hash_algo()
            except:
                h = hash_algo(usedforsecurity=False)  # Fedora/RHEL-specific

            if self.FILE in self._representations:
                if do_salt:
                    try:
                        salt = str(int(stat(self.FILE()).st_mtime))
                    except (OSError, TypeError):  # TypeError~coerce to Unicode
                        pass
                self.BYTESTRING()  # ensure bytestring repr (assumed "cheap")

            if not salt and do_salt:
                salt = str(int(time()))
            h.update(bytes_enc(salt))

            if self.BYTESTRING in self._representations:
                content = self.BYTESTRING()
            else:
                content = str(hash(self))
            h.update(bytes_enc(content))

            # only use first quarter of the whole hexdigest
            self._hash = h.hexdigest()[:h.digest_size//2]  # expected even
        return self._hash

    @classmethod
    def io_decl_specials(cls, io_decl, in_mode, magic_fds, interpolations={}):
        """Special file decl. treatment: "magic files", string interpolation

        Only return number if we've hit "magic file" (otherwise None),
        preceded with the string interpolation (via string.format())
        behind the curtains for "file" decl.
        """
        # turning @DIGIT+ magic files into fileobjs (needs global view)
        if tuplist(io_decl) and len(io_decl) >= 2 and io_decl[0] == cls.FILE:
            try:
                # incl. late/dynamic interpolation of defaults ~ filters' inputs
                io_decl = args2tuple(io_decl[0],
                                     io_decl[1].format(**interpolations),
                                     *io_decl[2:])
            except (ValueError, KeyError):
                pass
            except ((AttributeError, ) if not in_mode else ()):
                pass  # input terminals have to be interpolation-complete by now
            fdef = io_decl[1]
            fdef = "@{0}".format(int(not(in_mode))) if fdef == '-' else fdef
            if (isinstance(fdef, basestring)
                    and fdef.rstrip('0123456789') == '@'):
                fd = int(fdef[1:])
                if fd not in magic_fds:
                    # XXX be careful about not duplicating? (especially '-')
                    #     maybe even nothing should be duplicated at all?
                    try:
                        magic_fds[fd] = fdopen(fd, 'r+b' if in_mode else 'ab')
                    except (OSError, IOError):
                        # keep untouched
                        pass
                    else:
                        return io_decl
                io_decl = args2sgpl(io_decl[0], magic_fds[fd],
                                    *io_decl[2:])
        return io_decl


class CompositeFormat(Format, MetaPlugin):
    """Quasi-format to stand in place of multiple formats at once

    It is intended to build on top of atomic formats (and only these,
    i.e., multi-level nesting is not supported as it doesn't bring
    any better handling anyway).

    Common format API is overridden so as to be performed per each
    contained/designated format in isolation, whereas the aggregated
    result is then returned.

    Note that the semantics implicitly require protocols prescribing
    the `CompositeFormat` instantiation to also become "composite",
    i.e., having form like ('composite', ('file', 'file')) instead
    of mere scalar like 'file' (IOW, whole declaration remains
    fully "typed").

    See also: Format, SimpleFormat
    """
    native_protocol = 'composite'  # to be overridden by per-instance one
                                   # XXX: hybridproperty?

    # XXX def __new__

    def _construct(self, protocol, *args, **kwargs):
        """Format constructor, i.e., object = concrete multiformat data

        Parameters:
            protocol    protocol, should match: ('composite', ('file', ...))
                        where 'file' is indeed variable
            args        each should represent arguments to be passed
                        into format instantiation for respective (order-wise)
                        protocol within the composite one
            kwargs      further keyword arguments; supported:
                        formats     iterable of format classes involved,
                                    matching the order of the "proper data
                                    part" within protocols
        """
        assert isinstance(protocol, tuple) and protocol \
               and protocol[0] == self.__class__.native_protocol, \
               "protocol has to be tuple initiated with {0} for {1}" \
               .format(self.__class__.native_protocol, self.__class__.__name__)
        formats = kwargs.pop('formats')  # has to be present
        # further checks
        assert len(protocol) > 1
        assert tuplist(protocol[1]) and len(protocol[1]) > 1
        assert tuplist(formats) and len(protocol[1]) == len(formats)
        assert len(args) == len(formats)
        assert all(p in f._protocols or p == 'native'
                   for (f, p) in zip(formats, protocol[1]))
        self._protocols[protocol] = lambda *_: args  # just to pass the assert
        self.native_protocol = (self.__class__.native_protocol,
                                tuple(f.native_protocol for f in formats))
        # instantiate particular designated formats
        self._designee = tuple(
            f(p, *args2tuple(a), **kwargs)
            for (f, p, a) in zip(formats, protocol[1], args)
        )
        super(CompositeFormat, self)._construct(protocol, *args, **kwargs)

    def __iter__(self):
        return iter(self._designee)

    def __getitem__(self, index):
        return self._designee[index]

    def produce(self, protocol, *args, **kwargs):
        """Called by implicit invocation to get data externalized"""
        if protocol == 'native':
            protocol = self.native_protocol

        assert tuplist(protocol) and len(protocol) > 1
        assert protocol[0] == self.__class__.native_protocol
        assert len(protocol[1]) == len(self._designee)
        args = args or ((),) * len(protocol[1])
        # should call produce instead
        return tuple(f.producer(p)(p, *a, **kwargs)
                     for f, p, a in zip(self._designee, protocol[1], args))

    @property
    def hash(self):
        """Compute hash trying to uniquely identify the format instance"""
        if self._hash is None:
            self._hash = hex(reduce(lambda a, b: a ^ int(b.hash, base=16),
                                    self._designee, 0))[2:]
        return self._hash  # XXX hex digest can possibly be shorter that normal


class XML(SimpleFormat):
    """Base for XML-based configuration formats"""

    MAX_DEPTH = 1000

    root = ''  # root tag of the XML document

    @staticmethod
    def _walk_schema_step_up(tree_stack):
        """Step up within the tree_stack (bottom-up return in dir structure)"""
        child_root, child_data = (lambda x, *y: (x, y))(*tree_stack.pop())
        log.debug("Moving upwards: `{0}' -> `{1}'"
                  .format(child_root, tree_stack[-1][0]))
        current_tracking = tree_stack[-1][2]
        if len(child_data[1]):
            name = basename(child_root)
            if name in current_tracking:
                to_update = current_tracking[name][1]
            else:
                to_update = current_tracking
            to_update.update(child_data[1])

        return current_tracking

    @staticmethod
    def _walk_schema_step_down(tree_stack, root):
        """Step down within the tree_stack (top-down diving in dir structure)

        Based on the fact that we traverse down by one level to already
        (shallowly) explored level so the item is already tracked at
        the parent and the shallow knowledge (dict) is passed down to be
        potentially extended -- as dict is mutable, this knowledge
        is shared between parent and child level.
        """
        log.debug("Moving downwards:`{0}' -> `{1}'"
                  .format(tree_stack[-1][0], root))
        tree_stack.append((root, None, {}))

        return tree_stack[-1][2]  # current tracking

    @classmethod
    def walk_schema(cls, root_dir, symbol=None, preprocess=lambda s, n: s,
                    sparse=True, xml_root=None):
        """
        Get recipe for visiting symbol(s) within the XML as (sparsely) arranged

        Example of output::

            {
                'A': (<symbol>, {
                    'C': (<symbol>, {}),
                    'D': (<symbol>, {
                        'F': (<symbol>, {
                        })
                    })
                })
                'Z': (<symbol>, {})
            }

        NB: order of keys really does not matter.
        """
        xml_root = xml_root or cls.root
        particular_namespace = '.'.join((cls.namespace, symbol or xml_root))
        result = {}
        tree_stack = [(root_dir, None, result)]  # for bottom-up reconstruction
        for root, dirs, files in walk(root_dir):
            try:
                dirs.remove('__pycache__')  # PY3
            except ValueError:
                pass
            # multi-step upwards and (followed by)/or single step downwards
            while commonprefix((root, tree_stack[-1][0])) != tree_stack[-1][0]:
                cls._walk_schema_step_up(tree_stack)
            if root != tree_stack[-1][0]:
                current_tracking = cls._walk_schema_step_down(tree_stack, root)
            else:
                assert root == root_dir
                # at root, we do not traverse to any other dir than `xml_root`
                foreach(lambda d: d != xml_root and dirs.remove(d), dirs[:])
                current_tracking = tree_stack[-1][2]
                files = []

            for i in dirs + files:
                name, ext = splitext(i)  # does not hurt even if it is a dir
                if name.startswith('_') or name.startswith('.') \
                or i in files and ext != extsep + 'py':
                    continue

                log.debug("Trying `{0}' at `{1}'".format(name, root))
                mfile, mpath, mdesc = find_module(name, [root])
                # need to obfuscate the name due to, e.g., "logging" clash
                mname = '.'.join((particular_namespace, 'walk_' + name))
                # suppress problems with missing parent in module hierarchy
                modules.setdefault(particular_namespace, modules[__name__])
                if mname in modules:
                    mod = modules[mname]
                    if hasattr(mod, '__path__') and mod.__path__[0] != mpath:
                        # XXX robust?
                        raise FormatError(cls, "`{0}' already present"
                                          .format(mname))
                else:
                    try:
                        mod = load_module(mname, mfile, mpath, mdesc)
                    except ImportError as e:
                        log.warning("Cannot load `{0}': {1}".format(mpath, e))
                        continue
                    finally:
                        if mfile:
                            mfile.close()

                available = set(dir(mod)) - set(dir(type(mod)))
                swag = None
                if not symbol or symbol in available:
                    swag = getattr(mod, symbol) if symbol else tuple(available)
                    swag = preprocess(swag, name)
                if swag is None and (sparse or i in files):
                    continue  # files are terminals anyway
                current_tracking[name] = (swag, {})

        for i in xrange(cls.MAX_DEPTH):
            try:
                if cls._walk_schema_step_up(tree_stack) is result:
                    return result
            except IndexError:
                raise FormatError(cls, "Format tree structure inconsistency"
                                       " detected")
        else:
            raise FormatError(cls, "INFLOOP detected")

    ###

    native_protocol = ETREE = Protocol('etree')
    BYTESTRING = SimpleFormat.BYTESTRING

    validator_specs = {
        ETREE: '*.rng'  # grab whatever you'll find (with backtrack)
    }

    @classmethod
    def _etree_rng_validator_specs(cls, root_dir=None, spec=None):
        if modules[cls.__module__].__file__ == __file__:
            spec = ()    # skip validating for plain XML
        else:
            if spec is None:
                _, spec = cls._validators[cls.ETREE]  # installed by meta-level
            # XXX use spec getter a was the intention before
            if not root_dir:
                root_dir = dirname(modules[cls.__module__].__file__)
            if sep not in spec:
                spec = join(root_dir, cls.root, spec)
            if any(filter(lambda c: c in spec, '?*')):
                globbed = glob(spec)
                spec = globbed or spec
            elif exists(spec):
                spec = args2tuple(spec)
            if not tuplist(spec):
                raise FormatPlainError("Cannot validate, no matching spec:"
                                       " `{0}'".format(spec))
        return spec

    @classmethod
    def etree_rng_validator_proper_specs(cls, **kwargs):
        """Return class-exclusive validating RNG files, name/version ordered"""
        specs = set(cls._etree_rng_validator_specs(
            **filterdict_keep(kwargs, 'root_dir', 'spec')
        ))
        ancestor_specs = set()
        for ancestor in cls._update_from:
            this_ancestor_specs = ancestor._etree_rng_validator_specs(
                **filterdict_keep(kwargs, 'root_dir', 'spec')
            )
            if this_ancestor_specs:
                ancestor_specs.update(this_ancestor_specs)
        specs.difference(ancestor_specs)
        return sorted(specs, key=lambda x: namever_partition(splitext(x)[0]))

    @classmethod
    def etree_rng_validator(cls, et, start=None, **kwargs):
        """RNG-validate `et` ElementTree with schemes as per `root_dir`+`spec`

        ... and, optionally, narrowed to `start`-defined grammar segment.

        The `spec` can either be relative (`root` is an initial path then)
        or absolute wildcard (as in shell  globbing) or specific name.
        All validators matching the specification will be tried in an
        alphabetical order, first success wins, overall validation fails
        otherwise.

        `start` is either plain tag or namespaced one using Clark's notation.

        Returns tuple of:
            iterable of validation errors
            path of the "master" validating schema (first good one/last used)
            snippet from "master" validating schema relevant to `start`
        """
        # XXX holds its private cache under cls._validation_cache
        try:
            spec = cls._etree_rng_validator_specs(
                **filterdict_pop(kwargs, 'root_dir', 'spec')
            )
        except FormatPlainError as e:
            return (((0, 0, str(e)), ), None, None)
        fatal, master, master_snippet = [], '', ''
        for s in reversed(sorted(spec)):
            fatal, master = [], s
            try:
                schema, rng = cls._validation_cache.get(s, (None, None))
            except AttributeError:
                setattr(cls, '_validation_cache', {})
                schema = None
            if schema is None:
                try:
                    schema = etree.parse(s, parser=etree_parser_safe)
                    rng = etree.RelaxNG(schema)
                    cls._validation_cache[s] = schema, rng
                except (etree.RelaxNGError, etree.XMLSyntaxError):
                    log.warning("Problem processing RNG file `{0}'".format(s))
                    continue
            if start is not None:
                schema = deepcopy(schema)
                target = rng_pivot(schema, start)
                rng = etree.RelaxNG(schema)
            else:
                target = rng_get_start(schema)[0]
            try:
                rng.assertValid(et)
            except etree.DocumentInvalid:
                log.warning("Invalid as per RNG file `{0}'".format(s))
                for entry in rng.error_log:
                    fatal.append((entry.line, entry.column, entry.message))
                master_snippet = etree.tostring(target, pretty_print=True)
            else:
                break
        else:
            log.warning("None of the validation attempts succeeded with"
                        " validator spec: {0}".format(', '.join(spec)))

        return fatal, master, master_snippet.strip()

    etree_validator = etree_rng_validator

    @SimpleFormat.producing(BYTESTRING, chained=True)
    def get_bytestring(self, *iodecl):
        # chained fallback
        return etree.tostring(self.ETREE(protect_safe=True),
                              encoding='UTF-8', pretty_print=True)

    @SimpleFormat.producing(ETREE, protect=True,
                            validator=etree_validator.__get__(1).__func__)
    def get_etree(self, *iodecl):
        return etree.fromstring(self.BYTESTRING(),
                                parser=etree_parser_safe).getroottree()



# -*- coding: UTF-8 -*-
# Copyright 2016 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"


ccs_revitalize = '''\
    <!-- omit failoverdomains with repeated name
         rgmanager/src/daemons/fo_domain.c:fod_get_domain:error #31
     -->
    <xsl:template match="failoverdomain[
                             preceding-sibling::failoverdomain[
                                 @name = current()/@name
                             ]
                         ]">
        <xsl:message>
            <xsl:value-of select="concat('WARNING: omitting failoverdomain',
                                         ' with repeated name (', @name, ')')"/>
        </xsl:message>
    </xsl:template>
    <!-- warn on empty restricted failoverdomains
         rgmanager/src/daemons/groups.c:consider_start:instead of start req.
         http://oss.clusterlabs.org/pipermail/users/2016-January/002176.html
     -->
    <xsl:template match="failoverdomain[
                             @restricted != ''
                             and
                             contains('123456789',
                                       substring(@restricted, 1, 1))
                             and
                             not(failoverdomainnode)
                         ]">
        <xsl:message>
            <xsl:value-of select="concat('WARNING: empty restricted failoverdomain',
                                         ' means associated resource groups',
                                         ' will not start (', @name, ')')"/>
        </xsl:message>
        <xsl:copy>
            <xsl:apply-templates select="@*|node()"/>
        </xsl:copy>
    </xsl:template>
'''



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""*2pcscmd filters helpers"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from xml.sax.saxutils import escape
from base64 import b64encode
from hashlib import sha256
from os import getpid
from time import time

from ..utils_2to3 import bytes_enc, str_enc
from ..utils_prog import getenv_namespaced
from ..utils_xslt import NL

verbose_prefix = ':: '
verbose_OK = verbose_prefix + 'OK'
verbose_FAILURE = verbose_prefix + 'FAILURE'

ec_test = "{0}{1}:{1}".format(
    escape("test $? -eq 0 && echo '{verbose_OK}' || echo '{verbose_FAILURE}'"
           .format(verbose_OK=verbose_OK, verbose_FAILURE=verbose_FAILURE),
           {"'": "&apos;", '"': "&quot;"}),
    NL
)

verbose_ec_test = '''\
    <xsl:if test="$pcscmd_verbose">
        <xsl:value-of select='"%(ec_test)s"'/>
    </xsl:if>
''' % dict(
    ec_test=ec_test
)

def verbose_inform(what):
    return '''\
    <xsl:if test="$pcscmd_verbose">
        <xsl:value-of select='concat("echo &apos;%(verbose_prefix)s",
                                     %(what)s,
                                     "&apos;%(NL)s")'/>
    </xsl:if>
''' % dict(
    what=what or '""',
    verbose_prefix=verbose_prefix or '""',
    NL=NL,
)


def coro2pcscmd(**kwargs):
    descent = lambda w: \
        '<clufter:descent-mix at="{what}"/>'.format(
            what=kwargs.get(w) or w
        ) if w in kwargs else ''
    return ('''\
    <xsl:variable name="ClusterName">
        <xsl:choose>
            <xsl:when test="string(@name|totem/@cluster_name)">
                <xsl:value-of select="string(@name|totem/@cluster_name)"/>
            </xsl:when>
            <xsl:otherwise>
                <xsl:value-of select="'%(randomized)s'"/>
            </xsl:otherwise>
        </xsl:choose>
    </xsl:variable>
    <xsl:if test="not($pcscmd_dryrun)">
        <xsl:if test="not($pcscmd_noauth)">
''' + (
            verbose_inform('"auth cluster: ", $ClusterName')
) + '''
            <xsl:value-of select="'pcs cluster auth'"/>

            %(descent_node)s
            <xsl:if test="$pcscmd_force">
                <xsl:value-of select="' --force'"/>
            </xsl:if>
            <xsl:value-of select="' &lt;&gt; /dev/tty'"/>
            <xsl:value-of select="'%(NL)s'"/>
''' + (
            verbose_ec_test
) + '''
        </xsl:if>
        <xsl:if test="not($pcscmd_noguidance)">
            <!-- see rhbz#1210833 -->
''' + (
            verbose_inform('"check cluster includes local machine: ", $ClusterName')
) + r'''
            <xsl:value-of select="concat(
                'for l in $(comm -12',
                    ' &lt;(python -m json.tool /var/lib/pcsd/pcs_users.conf',
                        ' | sed -n &quot;s|^\s*\&quot;[^\&quot;]\+\&quot;:\s*\&quot;\([0-9a-f-]\+\)\&quot;.*|\1|1p&quot;',
                        ' | sort)',
                    ' &lt;(python -m json.tool /var/lib/pcsd/tokens',
                        ' | sed -n &quot;s|^\s*\&quot;[^\&quot;]\+\&quot;:\s*\&quot;\([0-9a-f-]\+\)\&quot;.*|\1|1p&quot;',
                        ' | sort)',
                ') @SENTINEL@; do',
                ' grep -Eq &quot;$(python -m json.tool /var/lib/pcsd/tokens',
                    ' | sed -n &quot;s|^\s*\&quot;\([^\&quot;]\+\)\&quot;:\s*\&quot;${l}\&quot;.*|\1|1p&quot;)&quot;',
                    ' - &lt;&lt;&lt;&quot;')"/>
            %(descent_node)s
            <xsl:value-of select="concat(
                '&quot; &amp;&amp; break;',
                ' false;',
                ' done || {',
                ' echo &quot;WARNING: cluster being created ought to include this very local machine&quot;;',
                ' read -p &quot;Do you want to continue [yN] (60s timeout): &quot; -t 60 || :;',
                ' test &quot;${REPLY}&quot; = &quot;y&quot; || kill -INT $$;',
                ' }%(NL)s'
            )"/>
        </xsl:if>
''' + (
        verbose_inform('"new cluster: ", $ClusterName')
) + '''
        <xsl:value-of select="'pcs cluster setup'"/>
        <xsl:choose>
            <xsl:when test="$pcscmd_enable">
                <xsl:value-of select="' --enable'"/>
            </xsl:when>
            <xsl:otherwise>
                <xsl:message>%(msg_enable)s</xsl:message>
            </xsl:otherwise>
        </xsl:choose>
        <xsl:value-of select="concat(' --name ', $ClusterName)"/>

        %(descent_node)s
        %(descent_cman)s
        %(descent_totem)s
        %(descent_quorum)s
        <xsl:value-of select="'%(NL)s'"/>

        <xsl:choose>
            <xsl:when test="$pcscmd_start_wait &gt; 0">
''' + (
                verbose_inform('"starting cluster and waiting for it to come up: ", $pcscmd_start_wait, " seconds"')
) + '''
            </xsl:when>
            <xsl:otherwise>
''' + (
                verbose_inform('"starting cluster"')
) + '''
            </xsl:otherwise>
        </xsl:choose>

        <xsl:value-of select="'pcs cluster start --all'"/>
        <xsl:if test="$pcscmd_start_wait &gt; 0">
            <xsl:choose>
                <xsl:when test="$pcscmd_extra_wait_cluster_start">
                    <xsl:value-of select="concat(' --wait=',
                                                 $pcscmd_start_wait)"/>
                </xsl:when>
                <xsl:otherwise>
                    <xsl:value-of select="concat(' &amp;&amp; sleep ',
                                                  $pcscmd_start_wait)"/>
                </xsl:otherwise>
            </xsl:choose>
            <xsl:value-of select="'%(NL)s'"/>
''' + (
            verbose_ec_test
) + '''
        </xsl:if>
    </xsl:if>
''') % dict(
    NL=NL,
    msg_enable="NOTE: cluster infrastructure services not enabled"
               " at this point, which can be changed any time by issuing"
               " `pcs cluster enable --all`",
    randomized='_NEEDED' + (
        '-' + str_enc(b64encode(sha256(
            bytes_enc(str(getpid()) + '_BY_PCS_' + str(int(time())), 'ascii')
        ).digest()), 'ascii')[0:8].replace('+', '-').replace('/', '_')
        if getenv_namespaced('NOSALT', '0') in ('0', 'false') else '_'
    ),
    **dict(('descent_' + k, descent(k))
           for k in ('cman', 'node', 'quorum', 'totem'))
)



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"


cibprelude2cibcompact = '''\
    <clufter:descent-mix preserve-rest="true"/>
    <!-- strip empty optional elements -->
    <xsl:template match="fencing-topology[count(*) = 0]"/>
'''

###

from .... import package_name

cibcompact2cib = ('''\
    <xsl:template match="crm_config[
                             not(
                                 following-sibling::resources/primitive[
                                     @class = 'stonith'
                                 ]|following-sibling::resources/primitive[
                                     preceding-sibling::template[
                                         @class = 'stonith'
                                         and
                                         @id = current()/@template
                                     ]
                                 ]
                             )
                         ]">
        <xsl:copy>
            <xsl:comment> %(note_stonith)s </xsl:comment>
            <xsl:message>%(note_stonith)s</xsl:message>
            <cluster_property_set id="CRMCONFIG-bootstrap">
                <nvpair id="CRMCONFIG-bootstrap-STONITH-ENABLED"
                        name="stonith-enabled"
                        value="false"/>
            </cluster_property_set>
        </xsl:copy>
    </xsl:template>

    <xsl:template match="constraints">
        <xsl:copy>
            <xsl:for-each select="../resources/template[
                                      @provider = '%(package_name)s'
                                      and
                                      @type = 'temporary-failoverdomain'
                                  ]">
                <xsl:variable name="FailoverDomain" select="."/>
                <xsl:for-each select="../template[
                                        @provider = '%(package_name)s'
                                        and
                                        @type = 'temporary-service'
                                        and
                                        meta_attributes/nvpair[
                                            @name = 'domain'
                                        ]/@value = $FailoverDomain/@id
                                    ]">
                    <xsl:variable name="Service" select="."/>
                    <xsl:variable name="Resources"
                                  select="../primitive[
                                              meta_attributes/nvpair[
                                                  @name = 'rgmanager-service'
                                                  and
                                                  @value = $Service/@id
                                              ]
                                          ]
                                          |
                                          ../group[
                                            @id = concat($Service/@id, '-GROUP')
                                          ]"/>
                    <xsl:for-each select="$Resources">
                        <xsl:variable name="Resource" select="."/>

                        <!-- rsc_location ~ failoverdomain/failoverdomainnodes -->

                        <xsl:if test="count(
                                          $FailoverDomain/meta_attributes/nvpair[
                                              starts-with(@name, 'failoverdomainnode-')
                                          ]
                                      ) != 0">
                            <rsc_location id="CONSTRAINT-LOCATION-{$Resource/@id}"
                                          rsc="{$Resource/@id}">
                                <xsl:comment
                                ><xsl:value-of select="concat(' mimic failoverdomain (',
                                                              substring-after($FailoverDomain/@id, 'FAILOVERDOMAIN-'),
                                                              ') for ', $Service/@id, ' ')"
                                /></xsl:comment>
                                <xsl:for-each select="$FailoverDomain/meta_attributes/nvpair[
                                                          starts-with(@name, 'failoverdomainnode-')
                                                      ]">
                                    <rule id="CONSTRAINT-LOCATION-{$Resource/@id}-{@value}">
                                        <xsl:attribute name="score">
                                            <xsl:choose>
                                                <xsl:when test="$FailoverDomain/meta_attributes/nvpair[
                                                                    @name = 'ordered'
                                                                ]/@value = 1
                                                                and
                                                                number(
                                                                    substring-after(@name,
                                                                                    'failoverdomainnode-')
                                                                ) &gt; 0">
                                                    <xsl:variable name="Result"
                                                                  select="(101 - number(
                                                                              substring-after(@name,
                                                                                              'failoverdomainnode-')
                                                                          )) * 10000"/>
                                                    <xsl:choose>
                                                        <xsl:when test="$Result = 1000000">
                                                            <xsl:value-of select="'INFINITY'"/>
                                                        </xsl:when>
                                                        <xsl:otherwise>
                                                            <xsl:value-of select="$Result"/>
                                                        </xsl:otherwise>
                                                    </xsl:choose>
                                                </xsl:when>
                                                <xsl:when test="$FailoverDomain/meta_attributes/nvpair[
                                                                    @name = 'ordered'
                                                                ]/@value = 1">
                                                    <xsl:value-of select="500000"/>
                                                </xsl:when>
                                                <xsl:otherwise>
                                                    <xsl:value-of select="'INFINITY'"/>
                                                </xsl:otherwise>
                                            </xsl:choose>
                                        </xsl:attribute>
                                        <expression id="CONSTRAINT-LOCATION-{$Resource/@id}-{@value}-expr"
                                                    attribute="#uname"
                                                    operation="eq"
                                                    value="{@value}"/>
                                    </rule>
                                </xsl:for-each>
                                <xsl:if test="$FailoverDomain/meta_attributes/nvpair[
                                                  @name = 'restricted'
                                              ]/@value = 1">
                                    <xsl:comment
                                    ><xsl:value-of select="concat(' mimic RESTRICTED failoverdomain (',
                                                                  substring-after($FailoverDomain/@id, 'FAILOVERDOMAIN-'),
                                                                  ') for ', $Service/@id, ' ')"
                                    /></xsl:comment>
                                    <rule id="CONSTRAINT-LOCATION-{$Resource/@id}-RESTRICTED"
                                          boolean-op="and"
                                          score="-INFINITY">
                                        <xsl:choose>
                                            <xsl:when test="$FailoverDomain/meta_attributes/nvpair[
                                                                starts-with(@name, 'failoverdomainnode-')
                                                            ]">
                                                <xsl:for-each select="$FailoverDomain/meta_attributes/nvpair[
                                                                          starts-with(@name, 'failoverdomainnode-')
                                                                      ]">
                                                    <expression id="CONSTRAINT-LOCATION-{$Resource/@id}-RESTRICTED-{@value}-expr"
                                                                attribute="#uname"
                                                                operation="ne"
                                                                value="{@value}"/>
                                                </xsl:for-each>
                                            </xsl:when>
                                            <xsl:otherwise>
                                                <!-- see filters/ccs-revitalize[failoverdomains]:
                                                     warn on empty restricted failoverdomains -->
                                                <expression id="CONSTRAINT-LOCATION-{$Resource/@id}-RESTRICTED-all-expr"
                                                            attribute="#uname"
                                                            operation="defined"/>
                                            </xsl:otherwise>
                                        </xsl:choose>
                                    </rule>
                                </xsl:if>
                            </rsc_location>
                        </xsl:if>
                    </xsl:for-each>
                    <xsl:if test="count($Resources) &gt; 1">
                        <rsc_order id="CONSTRAINT-ORDER-{$Service/@id}"
                                score="INFINITY">
                            <resource_set id="CONSTRAINT-ORDER-{$Service/@id}-RESOURCES">
                                <xsl:for-each select="$Resources">
                                    <resource_ref id="{@id}"/>
                                </xsl:for-each>
                            </resource_set>
                        </rsc_order>
                    </xsl:if>
                </xsl:for-each>
            </xsl:for-each>
        </xsl:copy>
    </xsl:template>
''') % dict(
    note_stonith="NOTE: no fencing is configured hence stonith is disabled;"
                 " please note, however, that this is suboptimal, especially"
                 " in shared storage scenarios",
    package_name=package_name()
)

###

from ....filters._2pcscmd import verbose_ec_test, verbose_inform
from ....utils_xslt import NL

cib2pcscmd = ('''\
    <xsl:if test="not($pcscmd_dryrun) and $pcscmd_tmpcib">
''' + (
        verbose_inform('"get initial/working CIB: ", $pcscmd_tmpcib')
) + '''
        <xsl:value-of select="concat('pcs cluster cib ',
                                     $pcscmd_tmpcib)"/>
        <xsl:value-of select="'%(NL)s'"/>
''' + (
        verbose_ec_test
) + '''
        <xsl:if test="$pcscmd_extra_push_diff">
''' + (
            verbose_inform('"store the original CIB for later comparison: ",'
                           ' $pcscmd_tmpcib, ".deltasrc"')
) + '''
            <xsl:value-of select="concat('cp ', $pcscmd_tmpcib,
                                         ' ', $pcscmd_tmpcib, '.deltasrc')"/>
            <xsl:value-of select="'%(NL)s'"/>
''' + (
            verbose_ec_test
) + '''
        </xsl:if>
    </xsl:if>
    <clufter:descent-mix at="crm_config"/>
    <clufter:descent-mix at="rsc_defaults"/>
    <clufter:descent-mix at="op_defaults"/>
    <clufter:descent-mix at="nodes"/>
    <clufter:descent-mix at="resources"/>
    <!-- constraints may refer to resource IDs from above -->
    <clufter:descent-mix at="constraints"/>
    <!-- fencing-level may refer to resource IDs from above,
         not enforced by schemas per se, but pcs mandates
         the referential integrity here, anyway
         (https://bugzilla.redhat.com/1441332#c7) -->
    <clufter:descent-mix at="fencing-topology"/>
    <clufter:descent-mix at="alerts"/>
    <!-- acls may refer to particular IDs (any) from above -->
    <clufter:descent-mix at="acls"/>

    <!-- enabling ACLs is sensible only at the very end of processing
         (was explicitly skipped while descenting into crm_config)
         XXX: should be more smoothly solved by using different modes
              with xsl:apply-templates, but that'd be quite an effort -->
    <xsl:variable name="CibAclEnable"
                  select="crm_config/cluster_property_set/nvpair[@name='enable-acl']"/>
    <xsl:if test="$CibAclEnable">


        <!-- "pcs acl" only supported with certain newer versions of
             pacemaker/pcs (https://bugzilla.redhat.com/1111369) -->
        <xsl:choose>
            <xsl:when test="$pcscmd_extra_acls">
''' + (
                verbose_inform('"set singleton cluster property: ", $CibAclEnable/@name')
) + '''
                <xsl:value-of select='concat($pcscmd_pcs, "property set")'/>
                <xsl:if test="$pcscmd_force">
                    <xsl:value-of select="' --force'"/>
                </xsl:if>
                <xsl:value-of select='concat(" &apos;", $CibAclEnable/@name,
                                             "=", $CibAclEnable/@value, "&apos;")'/>
                <xsl:value-of select="'%(NL)s'"/>
''' + (
                verbose_ec_test
) + '''
            </xsl:when>
            <xsl:otherwise>
                <xsl:message>%(acls_msg)s</xsl:message>
            </xsl:otherwise>
        </xsl:choose>
    </xsl:if>

    <xsl:if test="not($pcscmd_dryrun) and $pcscmd_tmpcib">
        <xsl:choose>
            <xsl:when test="$pcscmd_extra_push_diff">
''' + (
                verbose_inform('"push CIB (sending just the delta): ", $pcscmd_tmpcib')
) + '''
                <xsl:value-of select="concat('pcs cluster cib-push ',
                                             $pcscmd_tmpcib, ' diff-against=',
                                             $pcscmd_tmpcib, '.deltasrc')"/>
            </xsl:when>
            <xsl:otherwise>
''' + (
                verbose_inform('"push CIB: ", $pcscmd_tmpcib')
) + '''
                <xsl:value-of select="concat('pcs cluster cib-push ',
                                             $pcscmd_tmpcib, ' --config')"/>
            </xsl:otherwise>
        </xsl:choose>
        <xsl:value-of select="'%(NL)s'"/>
''' + (
        verbose_ec_test
) + '''
    </xsl:if>
''') % dict(
    NL=NL,
    acls_msg="WARNING: target pacemaker/pcs versions do not support"
             " (new) ACLs, hence &apos;enable-acl&apos; property omitted",
)

###

from ....utils_xslt import xslt_is_member, xslt_string_mapping

from itertools import chain

cib_revitalize_deprecated_props_cluster_rsc = {
    'default-resource-stickiness': 'resource-stickiness',
    'is-managed-default':          'is-managed',
}

cib_revitalize_deprecated_props_cluster_op = {
    'default-action-timeout': 'timeout',
}

cib_revitalize_deprecated_props_cluster = \
    chain(cib_revitalize_deprecated_props_cluster_rsc,
          cib_revitalize_deprecated_props_cluster_op)

cib_revitalize = ('''\
    <xsl:copy>

    <!-- nested handling of crm_config: drop the innovated nvpairs -->
    <clufter:descent-mix at="crm_config"/>

    <!-- move deprecated crm_config properties to rsc_defaults -->
    <xsl:choose>
        <xsl:when test="crm_config/cluster_property_set/nvpair[
''' + (
                             xslt_is_member('@name',
                                            cib_revitalize_deprecated_props_cluster_rsc)
) + ''']">
            <rsc_defaults>
                <xsl:apply-templates select="rsc_defaults/nvpair"/>
                <xsl:for-each select="crm_config/cluster_property_set[nvpair[
''' + (
                             xslt_is_member('@name',
                                            cib_revitalize_deprecated_props_cluster_rsc)
) + ''']]">
                    <meta_attributes id="{concat(@id, '-', generate-id(.))}">
                    <!-- XXX make rewrite-id a generalized helper -->
                    <!-- xsl:if test="rule">
                        <xsl:call-template name="rewrite-id">
                            <xsl:with-param name="Elem" select="rule"/>
                            <xsl:with-param name="InstanceId" select="generate-id(rule)"/>
                        </xsl:call-template>
                    </xsl:if -->
                    <xsl:for-each select="nvpair[
''' + (
                             xslt_is_member('@name',
                                            cib_revitalize_deprecated_props_cluster_rsc)
) + ''']">
                        <nvpair id="{concat(@id, '-', generate-id(.))}">
                            <xsl:attribute name="name">
                                <xsl:choose>
''' + (
                                xslt_string_mapping(cib_revitalize_deprecated_props_cluster_rsc, '@name')
) + '''
                                </xsl:choose>
                            </xsl:attribute>
                            <xsl:attribute name="value">
                                <xsl:value-of select="@value"/>
                            </xsl:attribute>
                        </nvpair>
                    </xsl:for-each>
                    </meta_attributes>
                    <xsl:apply-templates select="rsc_defaults/score"/>
                </xsl:for-each>
            </rsc_defaults>
        </xsl:when>
        <xsl:otherwise>
            <xsl:apply-templates select="rsc_defaults"/>
        </xsl:otherwise>
    </xsl:choose>

    <!-- move deprecated crm_config properties to op_defaults -->
    <xsl:choose>
        <xsl:when test="crm_config/cluster_property_set/nvpair[
''' + (
                             xslt_is_member('@name',
                                            cib_revitalize_deprecated_props_cluster_op)
) + ''']">
            <op_defaults>
                <xsl:apply-templates select="op_defaults/nvpair"/>
                <xsl:for-each select="crm_config/cluster_property_set[nvpair[
''' + (
                             xslt_is_member('@name',
                                            cib_revitalize_deprecated_props_cluster_op)
) + ''']]">
                    <meta_attributes id="{concat(@id, '-', generate-id(.))}">
                    <!-- XXX make rewrite-id a generalized helper -->
                    <!-- xsl:if test="rule">
                        <xsl:call-template name="rewrite-id">
                            <xsl:with-param name="Elem" select="rule"/>
                            <xsl:with-param name="InstanceId" select="generate-id(rule)"/>
                        </xsl:call-template>
                    </xsl:if -->
                    <xsl:for-each select="nvpair[
''' + (
                             xslt_is_member('@name',
                                            cib_revitalize_deprecated_props_cluster_op)
) + ''']">
                        <nvpair id="{concat(@id, generate-id(.))}">
                            <xsl:attribute name="name">
                                <xsl:choose>
''' + (
                                xslt_string_mapping(cib_revitalize_deprecated_props_cluster_op)
) + '''
                                </xsl:choose>
                            </xsl:attribute>
                            <xsl:attribute name="value">
                                <xsl:value-of select="."/>
                            </xsl:attribute>
                        </nvpair>
                    </xsl:for-each>
                    </meta_attributes>
                    <xsl:apply-templates select="op_defaults/score"/>
                </xsl:for-each>
            </op_defaults>
        </xsl:when>
        <xsl:otherwise>
            <xsl:apply-templates select="op_defaults"/>
        </xsl:otherwise>
    </xsl:choose>

    <xsl:apply-templates select="*[not(
''' + (
        xslt_is_member('name()',
                       ('crm_config', 'rsc_defaults', 'op_defaults'))
) + ''')]"/>

    </xsl:copy>
''')



# -*- coding: UTF-8 -*-
# Copyright 2014 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""Corosync executive configuration, XML representations"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..format import XML


class coroxml(XML):
    """Corosync executive configuration, XML version (corosync.xml)

    See corosync.xml(5).
    """
    # XMLFormat
    root = 'corosync'
    validator_specs = {
        XML.ETREE: ''  # XXX no RNG schema handy yet
    }


class coroxml_needle(coroxml):
    """Corosync (v2. needle) executive configuration, XML version"""
    validator_specs = {
        XML.ETREE: 'corosync.rng'
    }



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)

from __future__ import print_function

"""Base filter stuff (metaclass, decorator, etc.)"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from copy import deepcopy
from functools import reduce
from itertools import dropwhile, islice
from logging import getLogger
from os import environ, isatty, stat
from os.path import dirname, join
from re import compile as re_compile
from shlex import split as shlex_split
from shutil import rmtree
from subprocess import CalledProcessError, check_call
from sys import modules, stderr, __stdin__
from tempfile import mkdtemp, NamedTemporaryFile
from time import time
from warnings import warn
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

from lxml import etree

from . import package_name
from .error import ClufterError, ClufterPlainError
from .format import CompositeFormat, Format, XML
from .plugin_registry import MetaPlugin, PluginRegistry
from .utils import args2tuple, arg2wrapped, \
                   filterdict_keep, filterdict_invkeep, filterdict_pop, \
                   head_tail, hybridproperty, \
                   identity, lazystring, tuplist
from .utils_2to3 import MimicMeta, basestring, \
                        bytes_enc, str_enc, \
                        iter_items, iter_values, \
                        filter_u, foreach_u, reduce_u, \
                        unicode, xrange
from .utils_lxml import etree_XSLT_safe, \
                        etree_parser_safe, etree_parser_safe_unblanking
from .utils_func import apply_preserving_depth, \
                        apply_aggregation_preserving_depth, \
                        apply_intercalate, \
                        foreach, \
                        loose_zip, \
                        zip_empty
from .utils_prog import FancyOutput, ProtectedDict, \
                        cli_decor, cli_undecor, docformat, which
from .utils_xml import CLUFTER_NS, XSL_NS, \
                       namespaced, nselem, squote, element_juggler, \
                       xml_get_root_pi, \
                       xmltag_get_namespace, xmltag_get_localname
from .utils_xslt import xslt_identity
from .command_context import CommandContext

try:
    from .defaults import EDITOR
except ImportError:
    EDITOR = ''
EDITOR = environ.get('EDITOR', EDITOR)

CMD_HELP_OPTSEP_COMMON =  'common options:'

log = getLogger(__name__)

DEFAULT_ROOT_DIR = join(dirname(__file__), 'filters')

# XXX: consult standard/books
_TOP_LEVEL_XSL = (
    'attribute-set',
    'import',
    'include',
    'key',
    'namespace-alias',
    'output',
    'param',
    'strip-space',
    'template',
    'variable',
)

# requires context implying "weak, not a strong/standalone expression"
_IMPLIED_TOP_LEVEL_XSL = (
    'when',
)
TOP_LEVEL_XSL = [namespaced(XSL_NS, e) for e in
                 _TOP_LEVEL_XSL + _IMPLIED_TOP_LEVEL_XSL]


class FilterError(ClufterError):
    pass


class FilterPlainError(ClufterPlainError, FilterError):
    pass


class filters(PluginRegistry):
    """Filter registry (to be used as a metaclass for filters)"""
    pass


class _Filter(object):
    """Base for filters performing the actual conversion

    Base principles:
        - protocols: string label denoting how to int-/externalize
        - create filter instance = pass particular formats,
          call = start conversion
    """

    @MimicMeta.staticmethod
    def _resolve_formats_composite(formats):
        # XXX should rather be implemented by CompositeFormat itself?
        composite_onthefly = \
            lambda protocol, *args, **kwargs: \
                CompositeFormat(protocol, *args, **dict(iter_items(kwargs),
                                                        **{'formats': formats}))
        # XXX currently instantiation only (no match for composite classes)
        composite_onthefly.as_instance = \
            lambda *decl_or_instance, **kwargs: \
                composite_onthefly(('composite', ('native', ) * len(decl_or_instance)),
                                   *(di('native') for di in decl_or_instance), **kwargs) \
                if decl_or_instance and isinstance(decl_or_instance[0], Format)\
                else composite_onthefly(*decl_or_instance, **kwargs)
        composite_onthefly.context = CompositeFormat.context
        return composite_onthefly

    @MimicMeta.classmethod
    def _resolve_formats(cls, formats):
        res_input = [cls.in_format, cls.out_format]
        res_output = apply_preserving_depth(formats.get)(res_input)
        if apply_aggregation_preserving_depth(all)(res_output):
            log.debug("Resolve at `{0}' filter: `{1}' -> {2}"
                      .format(cls.name, repr(res_input), repr(res_output)))
            # capture composite formats if present;  when running
            # into composite format, we replace in-situ the whole iterable
            # with as-of-now resolved formats with lazily pulled
            # CompositeFormat passing it these formats along the standard
            # business (as opposed to on-the-fly class creation when it
            # probably won't be ever instantiated anyway);
            # extra lambda wrapping so as to surely make a closure around
            # ("remember correctly") the current value of res_output
            res_output = tuple(
                cls._resolve_formats_composite(o) if tuplist(o) else o
                for o in res_output
            )
            return res_output
        # drop the filter if cannot resolve any of the formats
        res_input = apply_intercalate(res_input)
        foreach_u(lambda i, x: log.warning("Resolve at `{0}' filter:"
                                           " `{1}' (#{2}) format fail"
                                           .format(cls.name, res_input[i], i)),
                  filter_u(lambda i, x: not(x),
                           enumerate(apply_intercalate(res_output))))
        return None

    @MimicMeta.method
    def __new__(cls, formats):
        io_formats = cls._resolve_formats(formats)
        if io_formats is None:
            return None
        self = super(Filter, cls).__new__(cls)
        (self._in_format, self._out_format), self._validated = io_formats, False
        return self

    @MimicMeta.passdeco(hybridproperty)
    def in_format(this):
        """Input format identifier/class for the filter"""
        return this._in_format

    @MimicMeta.passdeco(hybridproperty)
    def out_format(this):
        """Output format identifier/class for the filter"""
        return this._out_format

    @MimicMeta.method
    def __call__(self, in_obj, flt_ctxt=None, **kws):
        """Default is to use a function decorated with `deco`"""
        fmt_kws = filterdict_pop(kws, *self.out_format.context)
        if kws:
            warn("{0}: do not pass extraneous keyword arguments when not"
                 " testing".format(self.__class__.__name__), RuntimeWarning)
        if flt_ctxt is None:  # when estranged (not under Command control)
            cmd_ctxt = CommandContext()
            flt_ctxt = cmd_ctxt.ensure_filter(self)
            # following only possibly without taint protection (this branch)
            foreach(lambda d:
                    flt_ctxt.parent.update(filterdict_invkeep(d, *flt_ctxt)),
                    (dict(self.defs), kws))
        fmt_kws = filterdict_keep(flt_ctxt, *self.out_format.context, **fmt_kws)
        outdecl = self._fnc(flt_ctxt, in_obj)
        outdecl_head, outdecl_tail = head_tail(outdecl)
        outdecl_tail = arg2wrapped(outdecl_tail)
        if self._validated:
            fmt_kws['validator_specs'] = {'': ''}
        return self.out_format(outdecl_head, *outdecl_tail, **fmt_kws)

    @MimicMeta.classmethod
    def deco(cls, in_format, out_format, defs=None):
        """Decorator as an easy factory of actual filters"""
        def deco_fnc(fnc):
            log.debug("Filter: deco for {0}".format(fnc))
            attrs = {
                '__module__': fnc.__module__,
                '__doc__': fnc.__doc__,
                '_in_format': in_format,
                '_out_format': out_format,
                '_fnc': staticmethod(fnc),
                'defs': property(lambda self: ProtectedDict(defs)),
            }
            # optimization: shorten type() -> new() -> probe
            ret = cls.probe(fnc.__name__, (cls, ), attrs)
            return ret
        return deco_fnc

    @MimicMeta.classmethod
    def ctxt_svc_output(cls, ctxt, msg, **kwargs):
        if 'svc_output' in ctxt:
            svc_output = ctxt['svc_output']
        else:
            svc_output = FancyOutput(f=stderr, prefix="[{0}] ")
        svc_output(msg, prefix_arg=cls.name, **kwargs)


Filter = MimicMeta('Filter', filters, _Filter)


def tag_log(s, elem):
    """Logging helper"""
    return s.format(elem.tag, ', '.join(':'.join(i) for i in elem.items()))


class XMLFilter(Filter, MetaPlugin):
    """Base for XML/XSLT traversal filters"""

    _in_format = _out_format = 'XML'

    _re_highlight = re_compile('(?P<lp>`)(?P<msg>[^`]*)(?P<rp>`)')

    @staticmethod
    @docformat(CMD_HELP_OPTSEP_COMMON)
    def command_common(cmd_ctxt,
                       nocheck=False,
                       batch=False,
                       editor=EDITOR,
                       raw=False,
                       _profile=False):
        """\
        {0}
            nocheck   do not validate any step (even if self-checks present)
            batch     do not interact (validation failure recovery, etc.)
            editor    customize editor to run (unused in batch mode)
            raw       do not care about pretty-printed output
            _profile  enable XSLT profiling (auxiliary files produced)
        """
        flt_ctxt = cmd_ctxt.filter()
        flt_ctxt.setdefault('validator_specs', {'': ''} if nocheck else {},
                            bypass=True)
        flt_ctxt.update(
            raw=raw,
            interactive=not(batch and isatty(__stdin__.fileno())),
            editor=editor,
            profile=_profile,
        )

    @staticmethod
    def _traverse(in_fmt, walk, et=None,
                  walk_default_first=None, walk_default=None,
                  preprocess=lambda s, n, r: s, proceed=lambda *x: x,
                  postprocess=lambda x: x[0] if len(x) == 1 else x):
        """Generic traverse through XML as per symbols within schema tree"""
        default = walk_default_first
        default = default if default is not None else walk_default


        default_sym = etree.XML('<clufter:snippet'
               ' xmlns:xsl="{0}"'
               ' xmlns:clufter="{1}">'
               ' {2}'
               ' </clufter:snippet>'.format(XSL_NS, CLUFTER_NS, default))

        tree_stack = [('', ((default_sym, None, 2), walk), OrderedDict())]
        skip_until = []
        if default is None:
            skip_until = [('start', tag) for tag in walk]
        et = et or in_fmt('etree')

        for context in etree.iterwalk(et, events=('start', 'end')):
            event, elem = context
            log.debug("Got: {0} {1}".format(event, elem.tag))
            if skip_until and (event, elem.tag) not in skip_until:
                continue
            log.debug("Not skipped: {0}".format(elem.tag))
            skip_until = ()  # reset skipping any time we get further
            if event == 'start':
                # going down
                log.debug(tag_log("Moving downwards: {0} ({1})", elem))
                if elem.tag in tree_stack[-1][1][1] or default is not None:
                    if elem.tag not in tree_stack[-1][1][1]:
                        log.debug("Pushed to use default for `{0}'"
                                  .format(elem.tag))
                        previous = tree_stack[-1][1][1].copy()
                        tree_stack[-1][1][1].clear()
                        tree_stack[-1][1][1][elem.tag] = (default, previous)
                    walk_new_sym, walk_new_rest = tree_stack[-1][1][1][elem.tag]
                    default = walk_default  # for the rest under first/root
                    walk_new_sym = preprocess(walk_new_sym, elem.tag,
                                              tree_stack[-1][1][0])
                    tree_stack[-1][1][1][elem.tag] = (walk_new_sym, walk_new_rest)
                    tree_stack.append((elem.tag, (walk_new_sym, walk_new_rest), OrderedDict()))
                    if walk_new_rest is {}:
                        # safe optimization
                        skip_until = [('end', elem.tag)]
                # XXX: optimization prunning, probably no good
                #else:
                #    skip_until = [('end', tag) for tag in tree_stack[-1][1][1]]
                #    skip_until = [('end', elem.tag)]
                #    log.debug("Skipping (A) until: {0}".format(skip_until))

            else:
                # going up
                log.debug(tag_log("Moving upwards: {0} ({1})", elem))
                log.debug("Expecting {0}".format(elem.tag))
                if elem.tag == tree_stack[-1][0]:
                    walk, children = tree_stack.pop()[1:3]
                    tree_stack[-1][2][elem] = proceed(walk[0], elem, children)
                    try:
                        log.debug("Proceeded {0}".format(
                                  etree.tostring(tree_stack[-1][2][elem], encoding='unicode').replace('\n', '')))
                    except AttributeError:
                        log.debug("Proceeded {0}".format(tree_stack[-1][2][elem]))

                # XXX: optimization prunning, probably no good
                #else:
                #    skip_until = [('end', tree_stack[-1][0])]
                #    log.debug("Skipping (C) until: {0}".format(skip_until))

        ret = tuple(iter_values(tree_stack[-1][2]))
        # XXX can be () in case of not finding anything, should we emit error?
        #     addendum: sometimes comments (top-level only?) will cause this
        return postprocess(ret)

    @classmethod
    def _try_edit(cls, res_snippet, schema_path, schema_snippet, msgs, cnt,
                  use_offset=True, editor='', **ignored):
        editor = editor.strip() or EDITOR
        pkg_name = package_name()
        message = [
            "{0} manual validation failure recovery due to (local positions):",
            "",
            "{1}",
            "",
            "On the basis of the two snippets enclosed in (XML PI) separators",
            "revealing the purpose, to-be-fixed invalid result and excerpt of",
            "{2}",
            "(validating schema), the situation can be resolved by either:",
            "",
            ". realizing and FIXING the issue (and possibly notifying {0}",
            "  maintainer): directly EDIT the respective snippet (upon",
            "  exiting the editor, the snippet will be inspected again)",
            ". OMITTING the snippet from the result, but be warned, this",
            "  may cause validation error at the level closer to the root",
            ". TERMINATING the whole conversion: EXIT without a modification",
            "  {3} more time(s) in row or DELETE everything right away",
            "",
            "+ forcing the snippet, empty or not, without validation:",
            "  TURN `:force-this=false` to `:force-this=true`",
            "+ forcing the whole block of snippets without validation:",
            "  CHANGE `force-block` attribute in the root element to `true`",
            "",
            "Hint: `--dump={4} --nocheck` to store a full copy",
        ]
        offset = len(message) + len(msgs) + 5 if use_offset else 0
        e = '\n  '.join(["[{0}:{1}] {2}".format(m[0] + offset, *m[1:])
                        for m in msgs])
        message = '  ' + '\n  '.join(message).format(pkg_name, e, schema_path,
                                                     cnt, cli_decor(cls.name))
        prompt = """\
<{0}-recovery force-block="false">

<?{0} message
{1}
?>

<?{0} EDIT-result-snippet-start:force-this=false?>
{2}
<?{0} EDIT-result-snippet-end?>

<?{0} NOEDIT-schema-snippet-start?>
{3}
<?{0} NOEDIT-schema-snippet-end?>

</{0}-recovery>
""".format(pkg_name, message, res_snippet, schema_snippet)
        tmpdir = mkdtemp(prefix=pkg_name)
        reply, force = '', ''
        try:
            tmp_name = ""
            tmp = NamedTemporaryFile(mode='wb', dir=tmpdir, suffix='.xml',
                                     delete=False)
            with tmp as tmpfile:
                tmpfile.write(bytes_enc(prompt, 'utf-8'))
                tmpfile.flush()
                tmp_name = tmp.name
            old_stat = stat(tmp_name)

            editor_args = shlex_split(editor) + [tmp_name]
            assert len(editor_args) >= 2
            editor_args[0] = which(editor_args[0])
            try:
                # pty.spawn doesn't work as nicely,
                # /dev/tty may not be present (with open('/dev/tty') as si)
                # and we decide whether to be interactive per
                # sys.__stdin__ anyway
                log.info("running `{0}'".format(' '.join(editor_args)))
                check_call(editor_args, stdin=__stdin__)
            except (CalledProcessError, IOError) as e:
                raise FilterError(cls, str(e))
            except OSError:
                raise FilterError(cls, "Editor `{0}' seems unavailable"
                                        .format(editor))
            new_stat = stat(tmp_name)
            if old_stat.st_size == new_stat.st_size \
                    and old_stat.st_mtime == new_stat.st_mtime:
                return None, force  # no change occurred
            # do not trust editors/sed/whatever to do a _real in-place_
            # modifications (sed definitely doesn't; see also
            # http://www.pixelbeat.org/docs/unix_file_replacement.html),
            # otherwise tmpfile.seek(0) would be enough
            with open(tmp_name, 'rb') as tmpfile:
                reply = tmpfile.read().strip()
        finally:
            rmtree(tmpdir)
        if not reply:
            return False, force  # terminating
        elems = []
        reply = etree.fromstring(reply, parser=etree_parser_safe)
        if reply.attrib.get('force-block', '').lower() == 'true':
            force = 'block'
        for root_pi in xml_get_root_pi(reply):
            if (root_pi.target == pkg_name
            and root_pi.text.strip().startswith('EDIT-result-snippet-start')):
                text = root_pi.text[len('EDIT-result-snippet-start') + 1:]
                attrs = dict(a.strip().split('=') for a in text.split(':'))
                if attrs.get('force-this', 'false').lower() == 'true':
                    force = force or 'this'
                for sibling in root_pi.itersiblings():
                    if isinstance(sibling, type(root_pi)):
                        break
                    elems.append(sibling)
                break
        return elems, force

    @classmethod
    def _xslt_get_validate_hook(cls, validator, interactive=True, **kws):
        assert validator is not None

        def validate_hook(ret):
            global_msgs, single_elem, pkg_name = [], False, package_name()
            pi_comment = pkg_name + '-comment'

            # figure out the target element, skip if not any suitable
            if ret:
                # "protected" comments have to be turned to something
                # else, here a processing instruction
                cl = ret.xpath("//clufter:comment",
                               namespaces={'clufter': CLUFTER_NS})
                for e in cl:
                    element_juggler.rebind(etree.PI(pi_comment,
                                                    etree.tostring(e)),
                                           element_juggler.grab(e))
                    element_juggler.drop(e)
                #to_check = (ret.getroot(), )
                root = ret.getroot()
                if root.tag == namespaced(CLUFTER_NS, "snippet"):
                    to_check = reversed(root)
                else:
                    to_check = (root, )
                    single_elem = True
            else:
                to_check = ()
            worklist = list(i for i in to_check
                            if xmltag_get_namespace(i.tag) != XSL_NS)
            use_offset = True
            while worklist:
                elem = worklist.pop()
                if elem is None:
                    use_offset = True
                    continue
                msgs, schema, schema_snippet = validator(elem, start=elem.tag)
                if not (msgs and schema and interactive):
                    global_msgs.extend(msgs)
                    break
                parent_pos = element_juggler.grab(elem)
                res_snippet = str_enc(etree.tostring(elem, pretty_print=True,
                                                     encoding='UTF-8'),
                                      'utf-8').strip()
                force = False
                for i in xrange(2, 0, -1):  # 2 subsequent NOOPs -> termination
                    try:
                        elems, force = cls._try_edit(res_snippet.strip(),
                                                     schema, schema_snippet,
                                                     msgs, i, use_offset, **kws)
                    except FilterError as e:
                        log.warning(str(e))
                        elems = ()
                    if elems is not None:  # active change
                        break
                else:
                    print("Opportunity to recover the invalid (intermediate)"
                          " result was repeatedly abandoned", file=stderr)
                    elems = False

                if not elems:
                    element_juggler.drop(elem)
                    if elems is False:
                        print("Terminating", file=stderr)
                        raise SystemExit
                elif single_elem:
                    assert len(elems) == 1
                else:
                    # positive change occurred (reverse due to insert)
                    elems = reversed(elems)

                worklist.append(None)
                for e in elems:
                    if not force:
                        worklist.append(e)  # ensure revalidation
                    element_juggler.rebind(e, parent_pos)
                if force == 'block':
                    return ret, ()  # validation for the whole block cancelled
                use_offset = False

            cl = ret.xpath("//processing-instruction('{0}')".format(pi_comment))
            for e in cl:
                # XXX could be done better?  (e.text.strip().join((' ', ) * 2))
                reverted = etree.fromstring(e.text, parser=etree_parser_safe)
                element_juggler.rebind(nselem(CLUFTER_NS, 'comment', *tuple(
                                              reverted if len(reverted) else
                                              args2tuple(reverted.text))),
                                       element_juggler.grab(e))
                element_juggler.drop(e)
            return ret, global_msgs
        return validate_hook

    def _xslt_get_atom_hook(self, validator_specs={}, **kws):
        validate_hook = None
        if issubclass(self._out_format, XML):
            # only when out format is XML-based, we can be quite sure the
            # resulting form/protocol is etree
            # XXX otherwise there would have to be a hint about it
            #     or alternatively it would be passed into the filters
            #     generically, which might hurt implicit laziness?
            spec = validator_specs.get('etree', validator_specs.get('', None))
            validator = self._out_format.validator('etree', spec=spec)
            if validator:
                validate_hook = self._xslt_get_validate_hook(validator, **kws)
                self._validated = True  # to avoid Format instance revalidation
        return (lambda ret, error_log=():
                    self._xslt_atom_hook(ret, error_log, validate_hook, **kws))

    @classmethod
    def _xslt_atom_hook(cls, ret, error_log, validate_hook=None,
                        svc_output=(lambda msg, **kws:
                                    Filter.ctxt_svc_output({}, msg, **kws)),
                        **ignored):
        fatal = []
        for entry in error_log:
            emsg = entry.message
            if (entry.domain == 22 and entry.type == 0 and entry.level == 2
                    and emsg == "unknown error"):  # bogus errors
                continue
            urgent = (ret is None and not(emsg.split(' ', 1)[0].isupper())
                      or entry.type != 0)
            emsg = emsg if not urgent else 'FATAL: ' + emsg
            emsg = cls._re_highlight.sub('\g<lp>|highlight:\g<msg>|\g<rp>',
                                         emsg)
            svc_output("|subheader:xslt:| {0}".format(emsg), urgent=urgent,
                       base=reduce_u(
                           lambda now, new, new_l:
                               now or (emsg.startswith(new) and new_l),
                           iter_items({'WARNING:': 'warning', 'NOTE:': 'note'}),
                           ''
                       ) or urgent and 'error')
            if urgent:
                fatal.append("XSLT: " + entry.message)
        if not fatal:
            if hasattr(ret, 'xslt_profile') and ret.xslt_profile:
                profile = etree.tostring(ret.xslt_profile, pretty_print=True)
                fn = 'xslt-profile-{0}-{1}.xml'.format(cls.name,
                                                       hex(int(time()))[2:])
                with open(fn, "a") as f:
                    f.write(profile)
                    svc_output("|subheader:xslt-profile:| |highlight:{0}|"
                               .format(fn))
                del ret.xslt_profile

            if validate_hook:
                ret, entries = validate_hook(ret)
                fatal.extend("RNG: " + ':'.join(args2tuple(str(e[0]), str(e[1]),
                                                           *e[2:]))
                             for e in entries)
        if fatal:
            raise FilterError(cls, "{0}".format(', '.join(fatal)))
        return ret

    @staticmethod
    def _xslt_preprocess(sym, name, parent=None):
        """Preprocessing of schema tree XSLT snippets to real (sub)templates

        If callable is observed instead of XSLT snippet, keep it untouched.
        Used by `proceed_xslt` and `get_template` methods (hence class-wide).

        ...also, turn xls:comment into clufter:comment form so as to preserve
        (hold off) emitting XML comments, that would eitherwise be dropped
        because of repeated XSLT processings (xsl:comment -> literal comment
                                              -> forgotten)
        """
        # in top-down manner
        if isinstance(sym, tuple):
            return sym  # already proceeded
        if isinstance(sym, lazystring):
            sym = str(sym)
        if isinstance(sym, basestring):
            log.debug("preprocessing {0}".format(sym))
            # XXX <xsl:output method="xml"
            # XXX memoize as a constant + deepcopy
            sym = ('<clufter:snippet'
                   ' xmlns:xsl="{0}"'
                   ' xmlns:clufter="{1}">'
                   ' {2}'
                   ' </clufter:snippet>'.format(XSL_NS, CLUFTER_NS, sym))
            ret = etree.XML(sym)
            hooks = OrderedDict()

            log.debug("walking {0}".format(etree.tostring(ret)))
            will_mix = 0  # whether any descent-mix observed
            for event, elem in etree.iterwalk(ret, events=('start', )):
                # XXX xpath/specific tag filter
                # register each recurse point at the tag required
                # so it can be utilized in bottom-up pairing (from
                # particular definitions to where it is expected)

                # not needed
                #if elem is ret:
                #    continue
                log.debug("Got {0}".format(elem.tag))
                if elem.tag in (namespaced(CLUFTER_NS, t) for t
                                in ('descent', 'descent-mix')):
                    up = elem
                    walk = []
                    while up != ret:
                        walk.append(up.getparent().index(up))
                        up = up.getparent()
                    #walk = reversed(tuple(walk))  # XXX reversed, dangerous?
                    walk.reverse()
                    walk = tuple(walk)
                    mix = elem.tag == namespaced(CLUFTER_NS, 'descent-mix')
                    mix += mix and \
                            elem.attrib.get('preserve-rest', "false") == "true"
                    will_mix = max(will_mix, mix)
                    at = elem.attrib.get('at', '*')
                    # XXX can be a|b|c
                    at_hooks = hooks.setdefault(at, [])
                    at_hooks.append((walk, mix))
                    if len(at_hooks) > 1:
                        msg = ("Ambiguous match for `{0}' tag ({1} vs {2})"
                               .format(at, walk, at_hooks[0]))
                        if not mix:
                            raise FilterError(None, msg)
                        log.info(msg)
                elif (elem.tag == namespaced(XSL_NS, 'comment')
                      and parent and not(parent[2])):
                    # in non-root, turn the comments into "protected" ones
                    element_juggler.rebind(
                        nselem(CLUFTER_NS, 'comment', elem.text, *tuple(elem)),
                        element_juggler.grab(elem)
                    )
                    element_juggler.drop(elem)

            if parent and parent[2] and '*' not in hooks:
                hooks['*'] = [((len(ret), ), 1)]
                ret.append(nselem(CLUFTER_NS, 'descent-mix',
                                  attrib={'preserve-rest':
                                          ('false', 'true')[parent[2] - 1]}))

            # do_mix decides whether the current sub-template will be
            # be applied and the result attached (0), or just merged
            # to the parent template (1 if not preserve-rest required,
            # 2 otherwise)
            do_mix = parent[1].get(
                         name, parent[1].get('*', [(None, None)])
                     )[0][1] if parent and parent[1] is not None else will_mix
            if do_mix is None:
                raise RuntimeError("Parent does not expect `{0}' nor wildcard"
                                   .format(name))
            if do_mix and do_mix < will_mix:
                raise RuntimeError("Parent does not want preserve-rest while"
                                   " child wants to")
            elif do_mix > 1 and will_mix and not parent:
                do_mix = 1

            # note that when do_mix, nested top_levels are actually propagated
            # back, which is the inverse of what we are doing here
            if parent and isinstance(parent[0], etree._Element) and (
                not do_mix or parent[1] is None):
                top = filter(lambda x: x.tag in TOP_LEVEL_XSL, parent[0])
                for e in top:
                    #print("at", etree.tostring(ret), "appending", etree.tostring(e))
                    ret.append(deepcopy(e))

            log.debug("do_mix {0}, hooks {1}".format(do_mix, hooks))
            return (ret, hooks, do_mix)
        elif callable(sym):
            return sym
        else:
            log.debug("preprocess XSLT traverse symbols: skipping {0}"
                      .format(name))
            return None

    @classmethod
    def proceed_xslt(cls, in_obj, xslt_atom_hook=lambda ret, err: ret, **kws):
        """Apply iteratively XSLT snippets as per the schema tree (walk)

        You should likely use `filter_proceed_xslt` wrapper instead.
        """
        # XXX postprocess: omitted as standard defines the only root element

        def proceed(transformer, elem, children, profile=False):
            # expect (xslt, hooks) in the former case
            return xslt_atom_hook(*do_proceed(transformer, elem, children, profile)
                                  if not callable(transformer)
                                  else transformer(elem, children))

        def _merge_previous(snippet, hooks, elem, children):
            # snippet, an original preprocessed "piece of template puzzle",
            # has some of its subelements substituted as per hooks that
            # together with elem traversal and children dict decides which
            # parts (of previously proceeded symbols) will be grabbed
            scheduled = OrderedDict()  # XXX to keep the law and order
            for _, c_elem in etree.iterwalk(elem, events=('start',)):
                if c_elem is elem:
                    continue
                if c_elem in children:
                    c_up = c_elem
                    while not c_up.tag in hooks and c_up.getparent() != elem:
                        c_up = c_up.getparent()
                    target_tag = c_up.tag if c_up.tag in hooks else '*'
                    if c_up.tag in hooks or '*' in hooks:
                        for h in hooks[target_tag]:
                            l = scheduled.setdefault(h, [])
                            l.append(children[c_elem].getroot())

            for (index_history, mix), substitutes in iter_items(scheduled):
                tag = reduce(lambda x, y: x[y], index_history, snippet)
                parent = tag.getparent()
                index = parent.index(tag)

                for s in substitutes:
                    #assert s.tag == namespaced(CLUFTER_NS, 'snippet')
                    log.debug("before extension: {0}".format(etree.tostring(s)))
                    if s.tag == namespaced(CLUFTER_NS, 'snippet'):
                        # only single root "detached" supported (first == last)
                        dst = parent
                        # cannot use dict.update(dict) because of losing order
                        for k in s.attrib:
                            dst.attrib[k] = s.attrib[k]
                        #dst[index:index] = s
                        tag.extend(s)
                    elif mix:
                        tag.extend(s)
                    else:
                        # required by obfuscate
                        tag.append(s)
                    log.debug("as extended contains: {0}".format(etree.tostring(tag)))

                at = tag.attrib.get('at', '*')
                if mix == 1 and at != '*':  #and elem.getparent() is None:
                    e = nselem(XSL_NS, 'apply-templates', select=".//{0}".format(at))
                    tag.append(e)
                elif mix == 2:
                    e = nselem(XSL_NS, 'copy')
                    e.append(nselem(XSL_NS, 'apply-templates', select="@*|node()"))
                    tag.append(e)

            cl = snippet.xpath("//clufter:descent|//clufter:descent-mix",
                                 namespaces={'clufter': CLUFTER_NS})
            # remove these remnants so cleanup_namespaces works well
            for e in cl:
                parent = e.getparent()
                index = parent.index(e)
                parent[index:index] = e.getchildren()
                e.getparent().remove(e)

        def do_proceed(xslt, elem, children, profile=False):
            # in bottom-up manner

            hooks, do_mix = xslt[1:]
            error_log = ()
            # something already "mixed", shortcut, if first "mix" copy+clear
            if not len(xslt[0]):
                assert do_mix
                return xslt[0].getroottree(), error_log

            snippet = deepcopy(xslt[0])  # for in-situ template manipulation

            if do_mix:
                xslt[0].clear()  # if we mix, it is only once

            _merge_previous(snippet, hooks, elem, children)

            # XSLT to either be performed (do_mix == 0) or remembered (>0)
            xslt_root = nselem(XSL_NS, 'stylesheet', version="1.0")
            # move top-level items directly to the stylesheet being built
            if do_mix:
                xslt_root.text = snippet.text
            for e in etree.ETXPath(
                "./comment()|" + '|./'.join(tag
                                        for tag in TOP_LEVEL_XSL)
               + '|descendant::' + namespaced(XSL_NS, 'template')
            )(snippet):
                # special case for variable, as it may be needed within template
                # immediately
                if xmltag_get_localname(e.tag) == 'variable':
                    xslt_root.append(deepcopy(e))
                else:
                    xslt_root.append(e)

            # if something still remains, we assume it is "template"
            if tuple(islice(dropwhile(lambda x: x.tag in TOP_LEVEL_XSL, snippet), 1)):
                log.debug("snippet0: {0}, {1}, {2}".format(do_mix, elem.tag, etree.tostring(snippet)))
                template = nselem(XSL_NS, 'template', match=elem.tag)
                if do_mix:
                    template.extend(snippet)
                else:
                    template.append(snippet)
                log.debug("template1: {0}".format(etree.tostring(template)))
                #snippet.append(template)
                ##else:
                ##    template = snippet
                # ^ XXX was extend
                xslt_root.append(template)
                #print("ee", etree.tostring(xslt_root))

            # append "identity" to preserve application
            if do_mix > 1 or elem.getparent() is None and not children:
            # and not do_mix:
                template = etree.XML(xslt_identity(''))
                xslt_root.append(template)

            #else:
            #    # we dont't apply if there is nothing local and not at root
            #    print("zdrham", elem.tag)
            #    return elem

            if do_mix and elem.getparent() is not None:
                # "mix/carry" case in which we postpone this XSLT execution
                # (presumably non-local) by enquing it to the parent's turn
                #ret = xslt_root.getroot()
                ret = etree.ElementTree(xslt_root)
            else:
                # "eager" case in which we perform the (presumably local)
                # XSLT execution immediately
                elem = etree.ElementTree(elem)  # XXX not getroottree?
                log.debug("Applying {0}, {1}".format(type(elem), etree.tostring(elem)))
                log.debug("Applying on {0}".format(etree.tostring(xslt_root)))
                xslt = etree_XSLT_safe(xslt_root)
                try:
                    ret = xslt(elem, profile_run=profile)
                except etree.XSLTApplyError as e:
                    error_log = e.error_log
                    ret = None
                else:
                    # following seems to carefully preserve space (depending on
                    # xsl:output)
                    #ret = etree.fromstring(str(xslt(elem))).getroottree()
                    log.debug("With result {0}".format(etree.tostring(ret)))
                    #etree.cleanup_namespaces(ret)
                    error_log = xslt.error_log
            return ret, error_log

        def postprocess(ret):
            """Postprocess the final result

            ...also, turn clufter:comment back into (now true) comment form
            """
            #log.debug("Applying postprocess onto {0}".format(etree.tostring(ret)))
            assert len(ret) == 1
            ret = ret[0]
            if ret.getroot().tag == namespaced(CLUFTER_NS, 'snippet'):
                ret = ret.getroot()[0]

            # any "protected" comments are turned into full-fledged ones now
            cl = ret.xpath("//clufter:comment",
                           namespaces={'clufter': CLUFTER_NS})
            for e in cl:
                element_juggler.rebind(etree.Comment(e.text),
                                       element_juggler.grab(e))
                element_juggler.drop(e)

            # XXX: ugly solution to get rid of the unneeded namespace
            # (cleanup_namespaces did not work here)
            ret = etree.fromstring(etree.tostring(ret),
                                   parser=etree_parser_safe)
            etree.cleanup_namespaces(ret)
            return ret

        if not kws.pop('textmode', False):
            kws.setdefault('postprocess', postprocess)
        if kws.pop('profile', False):
            kws['proceed'] = (lambda *args, **kwargs:
                                  proceed(*args, profile=True, **kwargs))
        defaults = dict(preprocess=cls._xslt_preprocess, proceed=proceed,
                        sparse=True)
        defaults.update(kws)
        return cls.proceed(in_obj, **defaults)

    # XXX missing descent-mix
    @classmethod
    def _xslt_template(cls, walk):
        """Generate (try to) complete XSLT template from the sparse snippets"""
        scheduled_walk = [walk]
        scheduled_subst = OrderedDict()
        ret = []
        while len(scheduled_walk):
            cur_walk = scheduled_walk.pop()
            for key, (transformer, children) in iter_items(cur_walk):
                scheduled_walk.append(children)
                if transformer is None or callable(transformer):
                    if callable(transformer):
                        log.warning("Cannot generate complete XSLT when"
                                    " callable present")
                    if key in scheduled_subst:
                        for tag in scheduled_subst.pop(key)[:]:
                            for child_tag in children:
                                l = scheduled_subst.setdefault(child_tag, [])
                                l.append(tag)
                    continue

                snippet = deepcopy(transformer[0])  # in-situ manipulation

                xslt_root = nselem(XSL_NS, 'stylesheet', version="1.0")
                # XXX at least for xls:template, consider whole deep tree
                top = filter(lambda x: x.tag in TOP_LEVEL_XSL, snippet)
                for e in top:
                    xslt_root.append(e)
                if len(snippet):
                    snippet.tag = namespaced(XSL_NS, 'template')
                    snippet.attrib['match'] = key
                    xslt_root.append(snippet)

                hooks = transformer[1]
                if key not in scheduled_subst:
                    if cur_walk is walk:
                        ret.append(xslt_root)
                    else:
                        raise FilterError(cls, "XSLT inconsistency 1")
                # in parallel: 1?
                if key in scheduled_subst:
                    for tag in scheduled_subst.pop(key):
                        e = nselem(XSL_NS, 'apply-templates',
                                   select=".//{0}".format(key))
                        parent = tag.getparent()
                        parent[parent.index(tag)] = e
                        ret[-1].append(snippet)
                # in parallel: 2?
                for target_tag, at_hooks in iter_items(hooks):
                    for index_history, mix in at_hooks:
                        tag = reduce(lambda x, y: x[y], index_history, snippet)
                        l = scheduled_subst.setdefault(target_tag, [])
                        l.append(tag)

        assert not len(scheduled_subst)  # XXX either fail or remove forcibly
        foreach(lambda x: etree.cleanup_namespaces(x), ret)

        return (lambda x: x[0] if len(x) == 1 else x)(ret)

    @classmethod
    def proceed(cls, in_obj, root_dir=None, **kwargs):
        """Push-button to be called from the filter itself"""
        if not root_dir:
            root_dir = dirname(modules[cls.__module__].__file__)
        kwargs.setdefault('symbol', cli_undecor(cls.name))
        walk = in_obj.walk_schema(root_dir, **filterdict_pop(kwargs,
                                                             'symbol',
                                                             'sparse',
                                                             'xml_root'))
        walk_transform = kwargs.pop('walk_transform', identity)
        walk = walk_transform(walk)
        return cls._traverse(in_obj, walk, **kwargs)

    def filter_proceed_xslt(self, in_obj, **kwargs):
        """Push-button to be called from the filter itself, with walk_default"""
        raw, textmode = kwargs.pop('raw', False), kwargs.get('textmode', False)
        system = kwargs.pop('system', '')
        system_extra = kwargs.pop('system_extra', ())
        def_first = kwargs.pop('def_first', '') + (
            '<xsl:param name="system" select="{0}"/>'.format(squote(system))
        )
        if system:
            # guarantee at least 3 extra params, so they can be relied upon
            for i, val in loose_zip(xrange(1,4), system_extra):
                val = val if val is not zip_empty else ''
                def_first += ('<xsl:param name="system_{0}" select="{1}"/>'
                              .format(str(i), squote(val)))
        if textmode:
            def_first += '<xsl:output method="text" encoding="UTF-8"/>'
            def_first += '<xsl:strip-space elements="*"/>'
            def_first += '<clufter:descent-mix preserve-rest="false"/>'
        else:
            def_first += '<clufter:descent-mix preserve-rest="true"/>'

        xslt_atom_hook = self._xslt_get_atom_hook(**filterdict_pop(kwargs,
            'editor', 'interactive', 'svc_output', 'validator_specs'
        ))

        kwargs.setdefault('walk_default_first', def_first)
        kwargs['xslt_atom_hook'] = xslt_atom_hook

        ret = self.proceed_xslt(in_obj, **kwargs)
        if not raw and not textmode:
            # <http://lxml.de/FAQ.html#
            #  why-doesn-t-the-pretty-print-option-reformat-my-xml-output>
            ret = etree.fromstring(etree.tostring(ret),
                                   parser=etree_parser_safe_unblanking)
        elif textmode:
            ret = bytes_enc(unicode(ret), 'utf-8')
        return ret

    def ctxt_proceed_xslt(self, ctxt, in_obj, **kwargs):
        """The same as `filter_proceed_xslt`, context-aware"""
        kwargs = filterdict_keep(ctxt,
            'profile', 'raw', 'system', 'system_extra',  # <- proceed_xslt
            'editor', 'interactive', 'validator_specs',  # <- atom_hook
            'root_dir', 'walk_transform', 'xml_root',    # <- generic `proceed`
            **kwargs
        )
        kwargs['svc_output'] = ctxt.ctxt_svc_output
        return self.filter_proceed_xslt(in_obj, **kwargs)

    @classmethod
    def deco_xslt(cls, in_format, out_format, **kwargs):
        def deco_cls(new_cls):
            fnc = lambda ctxt, in_obj, **kwargsi: \
                      ('etree', ctxt.ctxt_proceed_xslt(in_obj,
                                                       **dict(kwargs,
                                                              **kwargsi)))
            fnc.__name__ = new_cls.__name__
            fnc.__module__ = new_cls.__module__
            return cls.deco(in_format, out_format)(fnc)
        return deco_cls

    @classmethod
    def get_template(cls, in_obj, root_dir=None, **kwargs):
        """Generate the overall XSLT template"""
        if not root_dir:
            root_dir = dirname(modules[cls.__module__].__file__)
        kwargs.setdefault('symbol', cli_undecor(cls.name))
        walk = in_obj.walk_schema(root_dir, preprocess=cls._xslt_preprocess,
                                  sparse=False,
                                  **filterdict_pop(kwargs, 'symbol',
                                                           'xml_root'))
        return cls._xslt_template(walk)



# -*- coding: UTF-8 -*-
# Copyright 2014 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""stringlist2stringset filter"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..filter import Filter

@Filter.deco('string-iter', 'string-set')
def stringlist2stringset(flt_ctxt, in_obj):
    """Downgrades list to a set (order not preserved)"""
    return ('stringset', set(in_obj('stringiter', protect_safe=True)))



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

# see also:
# https://github.com/ClusterLabs/pacemaker/blob/Pacemaker-1.1.17-rc2/xml/upgrade-1.3.xsl
# https://github.com/ClusterLabs/pacemaker/pull/1275
#
# NOTE: the first template is a local extension so as to explicitly set the
#       minimum schema version
fmt_cib_1to2 = '''\
<xsl:template match="cib">
  <xsl:copy>
    <xsl:apply-templates select="@*"/>
    <xsl:attribute name="validate-with">
      <xsl:value-of select="$cib2_min_ver"/>
    </xsl:attribute>
    <xsl:apply-templates select="node()"/>
  </xsl:copy>
</xsl:template>

<xsl:template match="role_ref">
  <xsl:element name="role">
    <xsl:apply-templates select="@*|node()"/>
  </xsl:element>
</xsl:template>

<xsl:template match="read|write|deny">
  <xsl:element name="acl_permission">

    <xsl:copy-of select="@id"/>
    <xsl:attribute name="kind"><xsl:value-of select="name()"/></xsl:attribute>

    <!-- previously, one could have a single element "matched" multiple times,
         each time using a different attribute (or no attribute at all), which
         would result, after the generalization (stripping @attribute) in
         multiple possibly conflicting ACL behaviours for given element(s);
         we could take this into account by, at the very least, preferring
         the behavior at attribute-less specification, if any -->
    <xsl:choose>
      <xsl:when test="@ref">
        <xsl:attribute name="reference"><xsl:value-of select="@ref"/></xsl:attribute>
        <xsl:if test="@attribute">
          <!-- alternatively, rephrase (generalized a bit) turning it to @xpath -->
          <xsl:message>ACLs: @attribute cannot accompany @ref for upgrade-1.3.xsl purposes, ignoring</xsl:message>
        </xsl:if>
      </xsl:when>
      <xsl:when test="@tag">
        <xsl:attribute name="object-type"><xsl:value-of select="@tag"/></xsl:attribute>
        <xsl:if test="@attribute">
          <xsl:message>ACLs: @attribute (with @tag) handling generalized a bit for upgrade-1.3.xsl purposes</xsl:message>
          <xsl:copy-of select="@attribute"/>
        </xsl:if>
      </xsl:when>
      <xsl:otherwise>
        <!-- must have been xpath per the schema, then -->
        <xsl:choose>
          <xsl:when test="@attribute">
            <xsl:message>ACLs: @attribute (with @xpath) handling generalized a bit for upgrade-1.3.xsl purposes</xsl:message>
            <xsl:attribute name="xpath">
              <xsl:value-of select="concat(@xpath,'[@', @attribute, ']')"/>
            </xsl:attribute>
          </xsl:when>
          <xsl:otherwise>
            <xsl:copy-of select="@xpath"/>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:otherwise>
    </xsl:choose>

  </xsl:element>
</xsl:template>

<xsl:template match="acl_user[role_ref]">
  <!-- schema disallows role_ref's AND deny/read/write -->
  <xsl:element name="acl_target">
    <xsl:apply-templates select="@*|node()"/>
  </xsl:element>
</xsl:template>

<xsl:template match="acl_user[not(role_ref)]">

  <xsl:element name="acl_target">
    <xsl:apply-templates select="@*"/>

    <xsl:if test="count(deny|read|write)" >
      <xsl:element name="role">
        <xsl:attribute name="id">
          <xsl:value-of select="concat('auto-', @id)"/>
        </xsl:attribute>
      </xsl:element>
    </xsl:if>

  </xsl:element>

  <xsl:if test="count(deny|read|write)" >
    <xsl:element name="acl_role">
      <xsl:attribute name="id">
        <xsl:value-of select="concat('auto-', @id)"/>
      </xsl:attribute>
      <xsl:apply-templates select="*"/>
    </xsl:element>
  </xsl:if>

</xsl:template>
'''



# -*- coding: UTF-8 -*-
version, alpha = '0.77.1', False

# https://www.python.org/dev/peps/pep-0440 + git export magic using export-subst
_git_hash = "98deec8".strip('$').replace("Format:%h", "")
_git_deco = ' (HEAD -> master, tag: v0.77.1, pg/master)'.strip('$()').replace("Format:%d", "")
_git_deco_arr = _git_deco.split(', ')
_git_tags = [i for i in _git_deco_arr if i.startswith("tag: v")]
_git_branches = [i for i in _git_deco_arr if i not in _git_tags + ['HEAD']]
if _git_branches and not _git_branches[-1].endswith('master') or alpha:
    if alpha:  # if not alpha, it is still not a true serving release
        version += 'a'  # no dashes(?)
    if _git_hash:
        version += '+git.{0}'.format(_git_hash)
elif _git_tags:
    assert any(t.endswith(version) for t in _git_tags), "version != tag"

license = 'GPLv2+'
copyright = """\
Copyright (C) 2018 Red Hat, Inc.
Licensed under {0}.
""".format(license).rstrip()

pkg_name = globals().get('__root__')

def package_name():
    global pkg_name
    if pkg_name is None:
        from os import readlink
        from os.path import abspath, basename, dirname, islink
        here = dirname(abspath(__file__))
        pkg_name = basename(abspath(readlink(here) if islink(here) else here))
    return pkg_name

# XXX eventually there should be precise plugin authorship tracking
author = ("Jan Pokorný <jpokorny+pkg-{0} @at@ Red Hat .dot. com>"
          " and plugin authors").format(package_name())

description = """\
Tool/library for transforming/analyzing cluster configuration formats

While primarily aimed at (CMAN,rgmanager)->(Corosync/CMAN,Pacemaker) cluster
stacks configuration conversion (as per RHEL trend), the command-filter-format
framework (capable of XSLT) offers also other uses through its plugin library.
"""
#Native plugins allows also for obfuscation of sensitive data within
#cluster configuration of supported cluster stacks, and other
#convenient actions.


_deobfuscate_email = (lambda what: what.replace(' @at@ ', '@')
                                       .replace(' .dot. ', '.')
                                       .replace('@Red Hat.', '@redhat.'))


def author_text(justname=None, first=False):
    if justname is not None or first:
        ret = ''
        authors = author.split(', ')
        first = first or len(authors) == 1
        for a in authors:
            name, rest = a.split('<', 1)
            email, rest = rest.split('>', 1)
            name, email, rest = map(str.strip, (name, email, rest))
            if justname:
                ret += ('' if first else ', ') + ' '.join((name, rest))
            else:
                ret += ('' if first else ', ') + _deobfuscate_email(email)
            if first:
                break
        return ret
    return _deobfuscate_email(author).lstrip('<').rstrip('>')


def copyright_text(justcopyright=None):
    if justcopyright is not None:
        c, licensed = copyright.split('\n', 1).rstrip()
        return c if justcopyright else licensed
    return copyright


def description_text(width=76, justhead=None):
    desc = description
    if justhead is False:
        desc = desc.split('\n', 2)[2]
    if not width or justhead is True:
        if justhead is True:
            desc = desc.split('\n', 1)[0]
            assert len(desc) <= (width or 1000)
    else:
        from textwrap import fill
        desc = map(lambda p: p.replace('\n', ' '), desc.split('\n\n'))
        desc = '\n\n'.join(map(lambda p: fill(p, width), desc))  #+ '\n'
    return desc


version_parts = (' '.join((package_name(), version)), copyright, '',
                 lambda: author_text().join(("Written by ", ".")))

def version_text(*args, **kwargs):
    args, sep = args + tuple(version_parts[len(args):]), kwargs.get('sep', '\n')
    return sep.join(a() if hasattr(a, '__call__') else a for a in args)



# -*- coding: UTF-8 -*-
# Copyright 2016 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""ccs2pcs{,-flatiron,-needle} commands"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..command import Command, CommandAlias
from ..facts import cluster_pcs_1_2, cluster_pcs_flatiron
from ..filter import XMLFilter
from ..protocol import protocols
from ..utils import args2tuple
from ..utils_cman import PATH_CLUSTERCONF
from ._chains_pcs import ccsflat2cibfinal_chain, ccsflat2cibfinal_output


def _check_pacemaker_1_2(cmd_ctxt):
    system = cmd_ctxt.get('system') or 'UNKNOWN-SYSTEM'
    system_extra = cmd_ctxt.get('system_extra') or ('UNKNOWN-DISTRO', )
    if not cluster_pcs_1_2(system, system_extra):
        from sys import stderr
        svc_output = cmd_ctxt.get('svc_output',
                                  lambda s, **kwargs: stderr.write(s + '\n'))
        svc_output("Resulting configuration will likely not be applicable to"
                   " ``{0}'' system as it seems so outdated as far as Pacemaker"
                   " not supporting validation schema v1.2"
                    .format(': '.join(args2tuple(system, *system_extra))),
                   base="error",
                   urgent=True,
        )


@Command.deco(('ccs2ccsflat',
                  ('ccs-disable-rg',
                      ('ccs2ccs-pcmk',
                          ('ccs-version-bump'))),
                  (ccsflat2cibfinal_chain)))
def ccs2pcs_flatiron(cmd_ctxt,
                     input=PATH_CLUSTERCONF,
                     ccs_pcmk="cluster-{ccs2ccsflat.in.hash}.conf",
                     cib="cib-{ccs2ccsflat.in.hash}.xml",
                     _common=XMLFilter.command_common):
    """(CMAN,rgmanager)->(Corosync/CMAN,Pacemaker) cluster cfg.

    More specifically, the output is suitable for Pacemaker integrated
    with Corosync ver. 1 (Flatiron) as present, e.g., in RHEL 6.{5, ..},
    and consists of Corosync/CMAN configuration incl. fencing pass-through
    (~cluster.conf) along with Pacemaker proper one (~cib.xml).

    Options:
        input     input (CMAN,rgmanager) cluster configuration file
        ccs_pcmk  output Corosync/CMAN (+fencing pass-through) config. file
        cib       output proper Pacemaker cluster config. file (CIB)
    """
    _check_pacemaker_1_2(cmd_ctxt)

    file_proto = protocols.plugins['file'].ensure_proto
    return (
        file_proto(input),
        (
            (
                (
                    file_proto(ccs_pcmk),
                ),
            ),
            ccsflat2cibfinal_output(
                file_proto(cib),
            ),
        ),
    )


@Command.deco(('ccs2ccsflat',
                  ('ccs-propagate-cman',
                      ('ccs2needlexml',
                          ('xml2simpleconfig'))),
                  (ccsflat2cibfinal_chain)))
def ccs2pcs_needle(cmd_ctxt,
                   input=PATH_CLUSTERCONF,
                   coro="corosync-{ccs2ccsflat.in.hash}.conf",
                   cib="cib-{ccs2ccsflat.in.hash}.xml",
                   _common=XMLFilter.command_common):
    """(CMAN,rgmanager)->(Corosync v2,Pacemaker) cluster cfg.

    More specifically, the output is suitable for Pacemaker integrated
    with Corosync ver. 2 (Needle) as present, e.g., in RHEL 7, and consists
    of Pacemaker (~cib.xml) and Corosync (~corosync.conf) configurations.

    Options:
        input     input (CMAN,rgmanager) cluster configuration file
        coro      output Corosync v2 config. file
        cib       output proper Pacemaker cluster config. file (CIB)
    """
    _check_pacemaker_1_2(cmd_ctxt)

    file_proto = protocols.plugins['file'].ensure_proto
    return (
        file_proto(input),
        (
            (
                (
                    file_proto(coro),
                ),
            ),
            ccsflat2cibfinal_output(
                file_proto(cib),
            ),
        ),
    )


@CommandAlias.deco
def ccs2pcs(cmds, *sys_id):
    # cluster_pcs_needle assumed unless "cluster_pcs_flatiron"
    return ccs2pcs_flatiron if cluster_pcs_flatiron(*sys_id) else ccs2pcs_needle



# -*- coding: UTF-8 -*-
# Copyright 2014 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)

from ....filters.ccs_artefacts import artefact

ccs_artefacts = artefact('@program', kind='C', desc='quorumd heuristic program')



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""cib2cibfinal filter"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..filter import XMLFilter


@XMLFilter.deco_xslt('cib', 'cib-final')
class cib2cibfinal: pass



# -*- coding: UTF-8 -*-
# Copyright 2014 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""pkgs2distropkgs filter"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..filter import Filter
from ..facts import package
from ..utils import args2sgpl


@Filter.deco('string-iter', 'string-iter')
def pkgs2distropkgs(flt_ctxt, in_obj):
    """Outputs distro packages based on 'logical packages'"""
    ret = args2sgpl(tuple(in_obj('native', protect_safe=True)))
    return ('native', (package(i, flt_ctxt['system'], flt_ctxt['system_extra'])
                       for i in ret if i))



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""Python 2-to-3 porting stuff (helpers, etc.)"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from collections import deque
from functools import reduce
from sys import modules, version_info


PY3 = version_info[0] >= 3


# Compatibility import chains (respective symbols to be used via this module)

try:
    StandardError = StandardError
except NameError:
    StandardError = Exception

try:
    basestring = basestring
except NameError:
    basestring = str

try:
    execfile = execfile
except NameError:
    def execfile(path, *globals_locals):
        with open(path) as f:
            source = compile(f.read(), f.name, 'exec')  # introspection-ready
            getattr(modules['builtins'], 'exec')(source, *globals_locals)

try:
    unicode = unicode
except NameError:
    unicode = str

try:
    xrange = xrange
except NameError:
    xrange = range


# Compatibility with dictionary methods not present in Python 3

# See https://www.python.org/dev/peps/pep-0469/#id9
try:
    dict.iteritems
except AttributeError:  # Python 3
    iter_values = lambda d: iter(d.values())
    iter_items = lambda d: iter(d.items())
else:  # Python 2
    iter_values = lambda d: d.itervalues()
    iter_items = lambda d: d.iteritems()


# Compatibility with implicit arguments unpacking not present in Python 3

# We were using implicit arguments unpacking heavily upon composite iterated
# elements coming as a second part of the `enumerate` product, especially in
# terse expressions using lambdas.  Python 3 does not allow this anymore,
# so we resort to using custom version of `enumerate` that will unpack the
# arguments for us.
# Similarly for: filter, map, reduce (incl. double nesting), and artificial
#                foreach function newly introduced in utils_func
enumerate_u = \
    lambda *seq_start: \
        ((lambda *a: a)(i, *item) for i, item in enumerate(*seq_start))

filter_u = lambda func, iterable: filter(lambda i: func(*i), iterable)
map_u = lambda func, *iterables: map(lambda i: func(*i), *iterables)
reduce_u = \
    lambda func, *sequence_initial: \
        reduce(lambda acc, new: func(acc, *new), *sequence_initial)
# double unpack, expects sequence of pairs of pairs
reduce_uu = \
    lambda func, *sequence_initial: \
        reduce(lambda acc, new: func(
            *(reduce(lambda a, rest: a + list(rest), (acc, new), []))
        ),*sequence_initial)

foreach_u = lambda *args: deque(map_u(*args), maxlen=0)


# Compatibility with bytes/string materialized dichotomy in Python 3

if PY3:
    bytes_enc = (lambda s, encoding='ascii':
                    bytes(s, encoding) if not isinstance(s, bytes) else s)
    str_enc = lambda s, encoding='ascii': str(s, encoding)
else:
    bytes_enc = lambda s, *args: bytes(s)
    str_enc = lambda s, *args: str(s)


# Compatibility with different way to use metaclasses in Python 2/3

class MimicMeta(object):
    """

    https://www.python.org/dev/peps/pep-3115/
    https://wiki.python.org/moin/PortingToPy3k/BilingualQuickRef#metaclasses
    """
    _sm = staticmethod  # workaround the identifiers-in-scope clash (deco)
    @_sm
    def method(m):
        setattr(m, '__mimicmeta__', None)
        return staticmethod(m)
    @_sm
    def staticmethod(m):
        setattr(m, '__mimicmeta__', staticmethod)
        return staticmethod(m)
    @_sm
    def classmethod(m):
        setattr(m, '__mimicmeta__', classmethod)
        return staticmethod(m)
    @_sm
    def passdeco(d):
        def _passdeco(m):
            setattr(m, '__mimicmeta__', d)
            return staticmethod(m)
        return _passdeco

    def __new__(cls, name, metaclass, deriveclass, bases=()):
        attrs = dict(__doc__=deriveclass.__doc__, __metaclass__=metaclass)
        for n in dir(deriveclass):
            attr = getattr(deriveclass, n)
            try:
                wrapper = attr.__mimicmeta__ or (lambda x: x)
                attrs.setdefault(n, wrapper(attr))
            except AttributeError:
                pass
        return metaclass(str(name), bases, attrs)
